from typing import List, Optional

from alitra import Pose
from pydantic import BaseModel, Field

from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.status import MissionStatus
from robot_interface.models.mission.task import TASKS
from robot_interface.utilities.uuid_string_factory import uuid4_string


class Mission(BaseModel):
    id: str = Field(default_factory=uuid4_string, frozen=True)
    tasks: List[TASKS] = Field(default_factory=list, frozen=True)
    name: str = Field(frozen=True)
    start_pose: Optional[Pose] = Field(default=None, frozen=True)
    dock: bool = Field(default=False, frozen=True)
    undock: bool = Field(default=False, frozen=True)
    status: MissionStatus = MissionStatus.NotStarted
    error_message: Optional[ErrorMessage] = Field(default=None)



import logging

from fastapi import Depends
from fastapi.security.base import SecurityBase
from fastapi_azure_auth import SingleTenantAzureAuthorizationCodeBearer
from fastapi_azure_auth.exceptions import InvalidAuth
from fastapi_azure_auth.user import User
from pydantic import BaseModel

from isar.config.settings import settings


class Token(BaseModel):
    access_token: str
    token_type: str


class NoSecurity(SecurityBase):
    def __init__(self) -> None:
        self.scheme_name = "No Security"


azure_scheme = SingleTenantAzureAuthorizationCodeBearer(
    app_client_id=settings.AZURE_CLIENT_ID,
    tenant_id=settings.AZURE_TENANT_ID,
    scopes={
        f"api://{settings.AZURE_CLIENT_ID}/user_impersonation": "user_impersonation",
    },
)


async def validate_has_role(user: User = Depends(azure_scheme)) -> None:
    """
    Validate if the user has the required role in order to access the API.
    Raises a 403 authorization error if not.
    """
    if settings.REQUIRED_ROLE not in user.roles:
        raise InvalidAuth(
            "Current user does not possess the required role for this endpoint"
        )


class Authenticator:
    def __init__(
        self,
        authentication_enabled: bool = settings.AUTHENTICATION_ENABLED,
    ) -> None:
        self.logger = logging.getLogger("api")
        self.authentication_enabled: bool = authentication_enabled
        enabled_string = "enabled" if self.authentication_enabled else "disabled"
        self.logger.info(f"API authentication is {enabled_string}")

    def should_authenticate(self):
        return self.authentication_enabled

    def get_scheme(self):
        if self.should_authenticate():
            return validate_has_role
        return NoSecurity

    async def load_config(self):
        """
        Load OpenID config on startup.
        """
        if self.should_authenticate():
            await azure_scheme.openid_config.load_config()
        else:
            pass



import json
import shutil
import time
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List

import numpy as np
import pytest
from alitra import Frame, Position
from fastapi import FastAPI
from injector import Injector
from starlette.testclient import TestClient

from isar.apis.api import API
from isar.config.settings import settings
from isar.modules import (
    APIModule,
    LocalPlannerModule,
    LocalStorageModule,
    QueuesModule,
    RequestHandlerModule,
    RobotModule,
    ServiceModule,
    StateMachineModule,
    SchedulingUtilitiesModule,
)
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import ReturnToHome
from tests.isar.state_machine.test_state_machine import (
    StateMachineThread,
    UploaderThread,
)
from tests.test_modules import MockMqttModule, MockNoAuthenticationModule


@pytest.fixture()
def injector_turtlebot():
    settings.ROBOT_PACKAGE = "isar_turtlebot"
    settings.LOCAL_STORAGE_PATH = "./tests/results"
    settings.PREDEFINED_MISSIONS_FOLDER = (
        "./tests/integration/turtlebot/config/missions"
    )
    settings.MAPS_FOLDER = "tests/integration/turtlebot/config/maps"
    settings.DEFAULT_MAP = "turtleworld"

    return Injector(
        [
            APIModule,
            MockNoAuthenticationModule,
            QueuesModule,
            RequestHandlerModule,
            RobotModule,
            ServiceModule,
            StateMachineModule,
            LocalPlannerModule,
            LocalStorageModule,
            SchedulingUtilitiesModule,
            MockMqttModule,
        ]
    )


@pytest.fixture()
def state_machine_thread(injector_turtlebot) -> StateMachineThread:
    return StateMachineThread(injector=injector_turtlebot)


@pytest.fixture()
def uploader_thread(injector_turtlebot) -> UploaderThread:
    return UploaderThread(injector=injector_turtlebot)


@pytest.fixture(autouse=True)
def run_before_and_after_tests() -> None:  # type: ignore
    results_folder: Path = Path("tests/results")
    yield

    print("Removing temporary results folder for testing")
    if results_folder.exists():
        shutil.rmtree(results_folder)
    print("Cleanup finished")


def test_successful_mission(
    injector_turtlebot, state_machine_thread, uploader_thread, access_token
) -> None:
    integration_test_timeout: timedelta = timedelta(minutes=5)
    app: FastAPI = injector_turtlebot.get(API).get_app()
    client: TestClient = TestClient(app=app)

    _id: int = 2

    _ = client.post(
        f"schedule/start-mission?ID={_id}",
        headers={"Authorization": "Bearer {}".format(access_token)},
    )
    time.sleep(5)

    mission_id: str = state_machine_thread.state_machine.current_mission.id
    mission: Mission = deepcopy(state_machine_thread.state_machine.current_mission)

    start_time: datetime = datetime.now(timezone.utc)
    while state_machine_thread.state_machine.current_state != States.Idle:
        if (datetime.now(timezone.utc) - start_time) > integration_test_timeout:
            raise TimeoutError
        time.sleep(5)

    mission_result_folder: Path = Path(f"tests/results/{mission_id}")

    drive_to_tasks: List[ReturnToHome] = [
        task for task in mission.tasks if isinstance(task, ReturnToHome)
    ]

    expected_positions: list = [task.pose for task in drive_to_tasks]

    paths = mission_result_folder.rglob("*.json")
    for path in paths:
        with open(path) as json_file:
            metadata = json.load(json_file)
        files_metadata: dict = metadata["data"][0]["files"][0]
        filename: str = files_metadata["file_name"]
        inspection_file: Path = mission_result_folder.joinpath(filename)

        actual_position: Position = Position(
            x=files_metadata["x"],
            y=files_metadata["y"],
            z=files_metadata["z"],
            frame=Frame("asset"),
        )

        close_to_expected_positions: List[bool] = []
        for expected_position in expected_positions:
            close_to_expected_positions.append(
                np.allclose(
                    actual_position.to_list(), expected_position.to_list(), atol=0.2
                )
            )

        assert any(close_to_expected_positions)
        assert inspection_file.exists()
    assert (
        len(list(mission_result_folder.glob("*"))) == 6
    )  # Expected two images, one thermal + three metadata files



import time
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field

from isar.apis.models.models import InputPose, InputPosition
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    DockingProcedure,
    Localize,
    RecordAudio,
    ReturnToHome,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
    TakeGasMeasurement,
    ZoomDescription,
)


class InspectionTypes(str, Enum):
    image = "Image"
    thermal_image = "ThermalImage"
    video = "Video"
    thermal_video = "ThermalVideo"
    audio = "Audio"
    gas_measurement = "GasMeasurement"


class TaskType(str, Enum):
    Inspection = "inspection"
    Localization = "localization"
    ReturnToHome = "return_to_home"
    Dock = "dock"


class StartMissionInspectionDefinition(BaseModel):
    type: InspectionTypes = Field(default=InspectionTypes.image)
    inspection_target: InputPosition
    analysis_type: Optional[str] = None
    duration: Optional[float] = None
    metadata: Optional[dict] = None


class StartMissionTaskDefinition(BaseModel):
    id: Optional[str] = None
    type: TaskType = Field(default=TaskType.Inspection)
    pose: InputPose
    inspection: Optional[StartMissionInspectionDefinition] = None
    tag: Optional[str] = None
    zoom: Optional[ZoomDescription] = None


class StartMissionDefinition(BaseModel):
    tasks: List[StartMissionTaskDefinition]
    name: Optional[str] = None
    start_pose: Optional[InputPose] = None
    dock: Optional[bool] = Field(default=False)
    undock: Optional[bool] = Field(default=False)


def to_isar_mission(
    start_mission_definition: StartMissionDefinition,
    return_pose: Optional[InputPose] = None,
) -> Mission:
    isar_tasks: List[TASKS] = []

    for task_definition in start_mission_definition.tasks:
        task: TASKS = to_isar_task(task_definition)
        isar_tasks.append(task)

    if return_pose:
        isar_tasks.append(ReturnToHome(pose=return_pose.to_alitra_pose()))

    if not isar_tasks:
        raise MissionPlannerError("Mission does not contain any valid tasks")

    isar_mission_name: str
    if start_mission_definition.name:
        isar_mission_name = start_mission_definition.name
    else:
        isar_mission_name = _build_mission_name()

    start_pose = None
    if start_mission_definition.start_pose:
        start_pose = start_mission_definition.start_pose.to_alitra_pose()

    return Mission(
        tasks=isar_tasks,
        name=isar_mission_name,
        start_pose=start_pose,
        dock=start_mission_definition.dock,
        undock=start_mission_definition.undock,
    )


def to_isar_task(task_definition: StartMissionTaskDefinition) -> TASKS:
    if task_definition.type == TaskType.Inspection:
        return to_inspection_task(task_definition)
    elif task_definition.type == TaskType.Localization:
        return to_localization_task(task_definition)
    elif task_definition.type == TaskType.ReturnToHome:
        return create_return_to_home_task(task_definition)
    elif task_definition.type == TaskType.Dock:
        return create_dock_task()
    else:
        raise MissionPlannerError(
            f"Failed to create task: '{task_definition.type}' is not a valid"
        )


def to_inspection_task(task_definition: StartMissionTaskDefinition) -> TASKS:
    inspection_definition = task_definition.inspection

    if inspection_definition.type == InspectionTypes.image:
        if task_definition.id:
            return TakeImage(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
        else:
            return TakeImage(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
    elif inspection_definition.type == InspectionTypes.video:
        if task_definition.id:
            return TakeVideo(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
        else:
            return TakeVideo(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
    elif inspection_definition.type == InspectionTypes.thermal_image:
        if task_definition.id:
            return TakeThermalImage(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
        else:
            return TakeThermalImage(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
    elif inspection_definition.type == InspectionTypes.thermal_video:
        if task_definition.id:
            return TakeThermalVideo(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
        else:
            return TakeThermalVideo(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
                zoom=task_definition.zoom,
            )
    elif inspection_definition.type == InspectionTypes.audio:
        if task_definition.id:
            return RecordAudio(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
            )
        else:
            return RecordAudio(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                target=task_definition.inspection.inspection_target.to_alitra_position(),
                duration=inspection_definition.duration,
                metadata=task_definition.inspection.metadata,
            )
    elif inspection_definition.type == InspectionTypes.gas_measurement:
        if task_definition.id:
            return TakeGasMeasurement(
                id=task_definition.id,
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                metadata=task_definition.inspection.metadata,
            )
        else:
            return TakeGasMeasurement(
                robot_pose=task_definition.pose.to_alitra_pose(),
                tag_id=task_definition.tag,
                metadata=task_definition.inspection.metadata,
            )
    else:
        raise ValueError(
            f"Inspection type '{inspection_definition.type}' not supported"
        )


def to_localization_task(task_definition: StartMissionTaskDefinition) -> Localize:
    return Localize(localization_pose=task_definition.pose.to_alitra_pose())


def create_return_to_home_task(
    task_definition: StartMissionTaskDefinition,
) -> ReturnToHome:
    return ReturnToHome(pose=task_definition.pose.to_alitra_pose())


def create_dock_task() -> DockingProcedure:
    return DockingProcedure(behavior="dock")


def _build_mission_name() -> str:
    return f"{settings.PLANT_SHORT_NAME}{settings.ROBOT_NAME}{int(time.time())}"



from http import HTTPStatus

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture

from isar.models.communication.queues.queues import QueueIO
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from tests.mocks.mission_definition import MockMissionDefinition


def test_timeout_send_command(
    mocker: MockerFixture, scheduling_utilities: SchedulingUtilities
):
    mocker.patch.object(QueueUtilities, "check_queue", side_effect=QueueTimeoutError)
    q: QueueIO = QueueIO(input_size=1, output_size=1)
    with pytest.raises(QueueTimeoutError):
        scheduling_utilities._send_command(True, q)
    assert q.input.empty()


def test_robot_capable_of_mission(scheduling_utilities: SchedulingUtilities):
    assert scheduling_utilities.verify_robot_capable_of_mission(
        mission=MockMissionDefinition.default_mission,
        robot_capabilities=["return_to_home", "take_image"],
    )


def test_robot_not_capable_of_mission(scheduling_utilities: SchedulingUtilities):
    with pytest.raises(HTTPException) as err:
        scheduling_utilities.verify_robot_capable_of_mission(
            mission=MockMissionDefinition.default_mission,
            robot_capabilities=["return_to_home"],
        )
    assert err.value.status_code == HTTPStatus.BAD_REQUEST


def test_state_machine_ready_to_receive_mission(
    scheduling_utilities: SchedulingUtilities,
):
    assert scheduling_utilities.verify_state_machine_ready_to_receive_mission(
        States.Idle
    )


def test_state_machine_not_ready_to_receive_mission(
    scheduling_utilities: SchedulingUtilities,
):
    with pytest.raises(HTTPException) as err:
        scheduling_utilities.verify_state_machine_ready_to_receive_mission(
            States.Initialize
        )
    assert err.value.status_code == HTTPStatus.CONFLICT



from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Type

from alitra import Pose
from pydantic import BaseModel, Field


@dataclass
class InspectionMetadata(ABC):
    start_time: datetime
    pose: Pose
    file_type: str
    analysis_type: Optional[str] = field(default=None, init=False)
    tag_id: Optional[str] = field(default=None, init=False)
    additional: Optional[dict] = field(default_factory=dict, init=False)


@dataclass
class ImageMetadata(InspectionMetadata):
    pass


@dataclass
class ThermalImageMetadata(InspectionMetadata):
    pass


@dataclass
class VideoMetadata(InspectionMetadata):
    duration: Optional[float] = field(default=None)


@dataclass
class ThermalVideoMetadata(InspectionMetadata):
    duration: Optional[float] = field(default=None)


@dataclass
class AudioMetadata(InspectionMetadata):
    duration: Optional[float] = field(default=None)


@dataclass
class GasMeasurementMetadata(InspectionMetadata):
    pass


class Inspection(BaseModel):
    metadata: InspectionMetadata
    id: str = Field(frozen=True)
    data: Optional[bytes] = Field(default=None, frozen=True)

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return InspectionMetadata


class Image(Inspection):
    metadata: ImageMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ImageMetadata


class ThermalImage(Inspection):
    metadata: ThermalImageMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ThermalImageMetadata


class Video(Inspection):
    metadata: VideoMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return VideoMetadata


class ThermalVideo(Inspection):
    metadata: ThermalVideoMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return ThermalVideoMetadata


class Audio(Inspection):
    metadata: AudioMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return AudioMetadata


class GasMeasurement(Inspection):
    metadata: GasMeasurementMetadata

    @staticmethod
    def get_metadata_type() -> Type[InspectionMetadata]:
        return GasMeasurementMetadata



import pytest
from fastapi.testclient import TestClient
from injector import Injector

from isar.apis.api import API
from isar.config.keyvault.keyvault_service import Keyvault
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.task_selector_interface import TaskSelectorInterface
from isar.models.communication.queues.queues import Queues
from isar.modules import (
    APIModule,
    LocalPlannerModule,
    QueuesModule,
    RequestHandlerModule,
    SequentialTaskSelectorModule,
    ServiceModule,
    StateMachineModule,
    SchedulingUtilitiesModule,
)
from isar.services.service_connections.request_handler import RequestHandler
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.idle import Idle
from isar.state_machine.states.initiate import Initiate
from isar.state_machine.states.monitor import Monitor
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from tests.mocks.robot_interface import MockRobot
from tests.test_modules import (
    MockAuthenticationModule,
    MockMqttModule,
    MockNoAuthenticationModule,
    MockRobotModule,
    MockStorageModule,
)


@pytest.fixture()
def injector():
    return Injector(
        [
            APIModule,
            LocalPlannerModule,
            MockMqttModule,
            MockNoAuthenticationModule,
            MockRobotModule,
            MockStorageModule,
            QueuesModule,
            RequestHandlerModule,
            ServiceModule,
            StateMachineModule,
            SequentialTaskSelectorModule,
            SchedulingUtilitiesModule,
        ]
    )


@pytest.fixture()
def injector_auth():
    return Injector(
        [
            APIModule,
            LocalPlannerModule,
            MockAuthenticationModule,
            MockMqttModule,
            MockRobotModule,
            MockStorageModule,
            QueuesModule,
            RequestHandlerModule,
            ServiceModule,
            StateMachineModule,
            SchedulingUtilitiesModule,
        ]
    )


@pytest.fixture()
def app(injector):
    app = injector.get(API).get_app()
    return app


@pytest.fixture()
def client(app):
    client = TestClient(app)
    return client


@pytest.fixture()
def client_auth(injector_auth):
    app = injector_auth.get(API).get_app()
    client_auth = TestClient(app)
    return client_auth


@pytest.fixture()
def access_token():
    return "DummyToken"


@pytest.fixture()
def keyvault(injector):
    return injector.get(Keyvault)


@pytest.fixture()
def state_machine(injector, robot):
    return StateMachine(
        queues=injector.get(Queues),
        robot=robot,
        mqtt_publisher=injector.get(MqttClientInterface),
        task_selector=injector.get(TaskSelectorInterface),
    )


@pytest.fixture()
def idle_state(state_machine):
    return Idle(state_machine)


@pytest.fixture()
def initiate(state_machine):
    return Initiate(state_machine)


@pytest.fixture()
def monitor(state_machine):
    return Monitor(state_machine=state_machine)


@pytest.fixture()
def request_handler(injector):
    return injector.get(RequestHandler)


@pytest.fixture()
def local_planner(injector):
    return injector.get(LocalPlanner)


@pytest.fixture()
def robot():
    return MockRobot()


@pytest.fixture()
def scheduling_utilities(app, injector):
    return injector.get(SchedulingUtilities)


@pytest.fixture()
def mission_reader(injector):
    return injector.get(LocalPlanner)



from datetime import datetime

from alitra import Frame, Orientation, Pose, Position

from robot_interface.models.inspection.inspection import ImageMetadata

MISSION_ID = "some-mission-id"
ARBITRARY_IMAGE_METADATA = ImageMetadata(
    start_time=datetime.now(),
    pose=Pose(
        Position(0, 0, 0, Frame("asset")),
        Orientation(0, 0, 0, 1, Frame("asset")),
        Frame("asset"),
    ),
    file_type="jpg",
)
DATA_BYTES: bytes = b"Lets say this is some image data"



import logging
from queue import Empty, Queue
from typing import Any

from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError

logger = logging.getLogger("api")


class QueueUtilities:
    """
    Contains utility functions for handling queue communication between threads.
    """

    @staticmethod
    def check_queue(queue: Queue, queue_timeout: int = None) -> Any:
        """
        Checks if there is a message on a queue. If a timeout is specified the function
        will raise a QueueTimeoutError if there is no message within the timeout. If
        there is no timeout specified this function will block.
        :param queue: The queue to be checked for a message
        :param queue_timeout: Timeout in seconds
        :return: Message found on queue
        :raises QueueTimeoutError
        """
        try:
            message: Any = queue.get(timeout=queue_timeout)
        except Empty:
            logger.error("Queue timed out")
            raise QueueTimeoutError
        return message

    @staticmethod
    def clear_queue(queue: Queue) -> None:
        while True:
            try:
                queue.get(block=False)
            except Empty:
                break



import logging
from typing import Any, Optional

import requests
from requests.exceptions import ConnectionError, HTTPError, RequestException, Timeout
from requests.models import Response

from isar.config.settings import settings


class RequestHandler:
    def __init__(self):
        self.logger = logging.getLogger("request_handler")

    def base_request(
        self,
        url: str,
        method: str,
        json_body: Any,
        timeout: float,
        auth: tuple,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        **kwargs,
    ) -> Response:
        try:
            response = requests.request(
                url=url,
                method=method,
                auth=auth,
                headers=headers,
                timeout=timeout,
                json=json_body,
                data=data,
                params=params,
                **kwargs,
            )
        except Timeout as e:
            self.logger.exception("Timeout exception")
            raise RequestException from e
        except ConnectionError as e:
            self.logger.exception("Connection error")
            raise RequestException from e
        except RequestException as e:
            raise RequestException from e
        except Exception as e:
            self.logger.exception("An unhandled exception occurred during a request")
            raise RequestException from e
        try:
            response.raise_for_status()
        except HTTPError:
            self.logger.exception(
                f"Http error. Http status code= {response.status_code}, Content: {response.content}"
            )
            raise
        return response

    def get(
        self,
        url: str,
        json_body=None,
        request_timeout: float = settings.REQUEST_TIMEOUT,
        auth: Optional[tuple] = None,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        **kwargs,
    ) -> Response:
        response = self.base_request(
            url=url,
            method="GET",
            auth=auth,
            headers=headers,
            timeout=request_timeout,
            json_body=json_body,
            data=data,
            params=params,
            **kwargs,
        )
        return response

    def post(
        self,
        url: str,
        json_body=None,
        request_timeout: float = settings.REQUEST_TIMEOUT,
        auth: Optional[tuple] = None,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        **kwargs,
    ) -> Response:
        response = self.base_request(
            url=url,
            method="POST",
            auth=auth,
            headers=headers,
            timeout=request_timeout,
            json_body=json_body,
            data=data,
            params=params,
            **kwargs,
        )
        return response

    def delete(
        self,
        url: str,
        json_body=None,
        request_timeout: float = settings.REQUEST_TIMEOUT,
        auth: Optional[tuple] = None,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        **kwargs,
    ) -> Response:
        response = self.base_request(
            url=url,
            method="DELETE",
            auth=auth,
            headers=headers,
            timeout=request_timeout,
            json_body=json_body,
            data=data,
            params=params,
            **kwargs,
        )
        return response

    def put(
        self,
        url: str,
        json_body=None,
        request_timeout: float = settings.REQUEST_TIMEOUT,
        auth: Optional[tuple] = None,
        headers: Optional[dict] = None,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        **kwargs,
    ) -> Response:
        response = self.base_request(
            url=url,
            method="PUT",
            auth=auth,
            headers=headers,
            timeout=request_timeout,
            json_body=json_body,
            data=data,
            params=params,
            **kwargs,
        )
        return response



import pytest

from isar.state_machine.states.monitor import Monitor
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, TaskStatus
from robot_interface.models.mission.task import ReturnToHome, TakeImage
from tests.mocks.task import MockTask


@pytest.mark.parametrize(
    "mock_status, expected_output",
    [
        (TaskStatus.Successful, True),
        (TaskStatus.Successful, True),
        (TaskStatus.Failed, True),
    ],
)
def test_task_finished(monitor: Monitor, mock_status, expected_output):
    task: ReturnToHome = MockTask.return_home()
    task.status = mock_status
    task_completed: bool = task.is_finished()
    assert task_completed == expected_output


@pytest.mark.parametrize(
    "is_status_successful, should_queue_upload",
    [
        (True, True),
        (False, False),
    ],
)
def test_should_only_upload_if_status_is_completed(
    monitor: Monitor, is_status_successful, should_queue_upload
):
    task: TakeImage = MockTask.take_image()
    task.status = TaskStatus.Successful if is_status_successful else TaskStatus.Failed
    mission: Mission = Mission(name="Dummy misson", tasks=[task])
    mission.status = (
        MissionStatus.Successful if is_status_successful else MissionStatus.Failed
    )

    monitor.state_machine.current_mission = mission
    monitor.state_machine.current_task = task

    if monitor._should_upload_inspections():
        monitor._queue_inspections_for_upload(mission, task)

    assert monitor.state_machine.queues.upload_queue.empty() == (
        not should_queue_upload
    )



from abc import ABCMeta, abstractmethod
from typing import Union

from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission


class StorageInterface(metaclass=ABCMeta):
    @abstractmethod
    def store(self, inspection: Inspection, mission: Mission) -> Union[str, dict]:
        """
        Parameters
        ----------
        mission : Mission
            Mission the inspection is a part of.
        inspection : Inspection
            The inspection object to be stored.

        Returns
        ----------
        String
            Path of the saved inspection

        Raises
        ----------
        StorageException
            An error occurred when storing the inspection.
        """
        pass


class StorageException(Exception):
    pass



import logging
from pathlib import Path
from typing import Union

from azure.core.exceptions import ResourceExistsError
from azure.storage.blob import BlobClient, BlobServiceClient, ContainerClient
from injector import inject

from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings
from isar.storage.storage_interface import StorageException, StorageInterface
from isar.storage.utilities import construct_metadata_file, construct_paths
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission


class BlobStorage(StorageInterface):
    @inject
    def __init__(
        self, keyvault: Keyvault, container_name: str = settings.BLOB_CONTAINER
    ):
        self.keyvault = keyvault
        self.storage_connection_string = self.keyvault.get_secret(
            "AZURE-STORAGE-CONNECTION-STRING"
        ).value
        self.container_name = container_name

        self.blob_service_client = self._get_blob_service_client()
        self.container_client = self._get_container_client(
            blob_service_client=self.blob_service_client
        )

        self.logger = logging.getLogger("uploader")

    def store(self, inspection: Inspection, mission: Mission) -> Union[str, dict]:
        data_path, metadata_path = construct_paths(
            inspection=inspection, mission=mission
        )

        metadata_bytes: bytes = construct_metadata_file(
            inspection=inspection, mission=mission, filename=data_path.name
        )

        self._upload_file(path=metadata_path, data=metadata_bytes)
        return self._upload_file(path=data_path, data=inspection.data)

    def _upload_file(self, path: Path, data: bytes) -> Union[str, dict]:
        blob_client = self._get_blob_client(path)
        try:
            blob_client.upload_blob(data=data)
        except ResourceExistsError as e:
            self.logger.error(
                f"Blob {path.as_posix()} already exists in container. Error: {e}"
            )
            raise StorageException from e
        except Exception as e:
            self.logger.error("An unexpected error occurred while uploading blob")
            raise StorageException from e

        absolute_inspection_path = {
            "source": "blob",
            "storage_account": settings.BLOB_STORAGE_ACCOUNT,
            "blob_container": settings.BLOB_CONTAINER,
            "blob_name": blob_client.blob_name,
        }
        return absolute_inspection_path

    def _get_blob_service_client(self) -> BlobServiceClient:
        try:
            return BlobServiceClient.from_connection_string(
                self.storage_connection_string
            )
        except Exception as e:
            self.logger.error(f"Unable to retrieve blob service client. Error: {e}")
            raise e

    def _get_container_client(
        self, blob_service_client: BlobServiceClient
    ) -> ContainerClient:
        return blob_service_client.get_container_client(self.container_name)

    def _get_blob_client(self, path_to_blob: Path) -> BlobClient:
        return self.container_client.get_blob_client(path_to_blob.as_posix())



from enum import Enum


class States(str, Enum):
    Off = "off"
    Idle = "idle"
    Initiate = "initiate"
    Initialize = "initialize"
    Monitor = "monitor"
    Paused = "paused"
    Stop = "stop"
    Offline = "offline"
    BlockedProtectiveStop = "blocked_protective_stop"

    def __repr__(self):
        return self.value



import json
import time
from datetime import datetime, timezone
from pathlib import Path
from typing import Tuple

from isar.config.settings import settings
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.utilities.json_service import EnhancedJSONEncoder


def construct_paths(inspection: Inspection, mission: Mission) -> Tuple[Path, Path]:
    folder: Path = Path(get_foldername(mission=mission))
    filename: str = get_filename(inspection=inspection)

    inspection_path: Path = folder.joinpath(
        f"{filename}.{inspection.metadata.file_type}"
    )

    metadata_path: Path = folder.joinpath(f"{filename}.json")

    return inspection_path, metadata_path


def construct_metadata_file(
    inspection: Inspection, mission: Mission, filename: str
) -> bytes:
    data: dict = {
        "coordinate_reference_system": settings.COORDINATE_REFERENCE_SYSTEM,
        "vertical_reference_system": settings.VERTICAL_REFERENCE_SYSTEM,
        "data_classification": settings.DATA_CLASSIFICATION,
        "source_url": None,
        "plant_code": settings.PLANT_CODE,
        "media_orientation_reference_system": settings.MEDIA_ORIENTATION_REFERENCE_SYSTEM,  # noqa: E501
        "additional_meta": {
            "inspection_id": inspection.id,
            "mission_id": mission.id,
            "mission_name": mission.name,
            "mission_date": datetime.now(timezone.utc).date(),
            "plant_name": settings.PLANT_NAME,
            "isar_id": settings.ISAR_ID,
            "robot_name": settings.ROBOT_NAME,
            "analysis_type": (
                inspection.metadata.additional["analysis_type"]
                if inspection.metadata.additional
                else "N/A"
            ),
        },
        "data": [
            {
                "folder": f"/{get_foldername(mission=mission)}",
                "files": [
                    {
                        "file_name": filename,
                        "timestamp": inspection.metadata.start_time,
                        "x": inspection.metadata.pose.position.x,
                        "y": inspection.metadata.pose.position.y,
                        "z": inspection.metadata.pose.position.z,
                        "tag": inspection.metadata.tag_id,
                        "additional_media_metadata": {
                            "orientation": inspection.metadata.pose.orientation.to_quat_array()  # noqa: E501
                        },
                    }
                ],
            }
        ],
    }

    return json.dumps(data, cls=EnhancedJSONEncoder, indent=4).encode()


def get_filename(inspection: Inspection) -> str:
    inspection_type: str = type(inspection).__name__
    tag: str = inspection.metadata.tag_id if inspection.metadata.tag_id else "no-tag"
    epoch_time: int = int(time.time())
    return f"{tag}__{inspection_type}__{epoch_time}"


def get_foldername(mission: Mission) -> str:
    mission_name: str = mission.name.replace(" ", "-")
    return f"{datetime.now(timezone.utc).date()}__{settings.PLANT_SHORT_NAME}__{mission_name}__{mission.id}"



import os
from importlib.resources import as_file, files
from typing import Any, List, Optional

from dotenv import load_dotenv
from pydantic import Field, ValidationInfo, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from isar.config import predefined_missions
from robot_interface.models.robots.robot_model import RobotModel
from robot_interface.telemetry.payloads import DocumentInfo


class Settings(BaseSettings):
    def __init__(self) -> None:
        try:
            source = files("isar").joinpath("config").joinpath("settings.env")
            with as_file(source) as eml:
                env_file = eml
        except ModuleNotFoundError:
            env_file = None
        super().__init__(_env_file=env_file)

    # Determines which robot package ISAR will attempt to import
    # Name must match with an installed python package in the local environment
    ROBOT_PACKAGE: str = Field(default="isar_robot")

    # Determines the local path in which results from missions are stored
    LOCAL_STORAGE_PATH: str = Field(default="./results")

    # Timeout in seconds for direct HTTP requests made through the RequestHandler
    REQUEST_TIMEOUT: int = Field(default=30)

    # Timeout in seconds for checking whether there is a message on a queue
    QUEUE_TIMEOUT: int = Field(default=10)

    # Sleep time for while loops in the finite state machine in seconds
    # The sleep is used to throttle the system on every iteration in the loop
    FSM_SLEEP_TIME: float = Field(default=0.1)

    # Location of JSON files containing predefined missions for the Local Planner to use
    path: str = os.path.dirname(predefined_missions.__file__)
    PREDEFINED_MISSIONS_FOLDER: str = Field(default=path + "/")

    # Name of default map transformation
    DEFAULT_MAP: str = Field(default="default_map")

    # Location of JSON files containing predefined maps
    MAPS_FOLDER: str = Field(default="src/isar/config/maps/")

    # Determines the number of state transitions that are kept in the log
    STATE_TRANSITIONS_LOG_LENGTH: int = Field(default=20)

    # Number of attempts to initiate a task or mission before cancelling
    INITIATE_FAILURE_COUNTER_LIMIT: int = Field(default=10)

    # Number of attempts to request a task status in monitor before cancelling
    REQUEST_STATUS_FAILURE_COUNTER_LIMIT: int = Field(default=3)

    # Time allocated to reconnect when failing to retrieve status due to communication
    # issues
    REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY: float = Field(default=10)

    # Number of attempts to stop the robot before giving up
    STOP_ROBOT_ATTEMPTS_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_ATTEMPTS_LIMIT: int = Field(default=10)

    # Number of attempts to stop the robot before giving up
    UPLOAD_FAILURE_MAX_WAIT: int = Field(default=60)

    # ISAR telemetry intervals
    ROBOT_STATUS_PUBLISH_INTERVAL: float = Field(default=1)
    ROBOT_HEARTBEAT_PUBLISH_INTERVAL: float = Field(default=1)
    ROBOT_INFO_PUBLISH_INTERVAL: float = Field(default=5)
    ROBOT_API_STATUS_POLL_INTERVAL: float = Field(default=5)

    # FastAPI host
    API_HOST_VIEWED_EXTERNALLY: str = Field(default="0.0.0.0")

    # FastAPI port
    API_PORT: int = Field(default=3000)

    # Determines which mission planner module is used by ISAR
    MISSION_PLANNER: str = Field(default="local")

    # Determines which task selector module is used by ISAR
    # Options: [sequential]
    TASK_SELECTOR: str = Field(default="sequential")

    # Determines which storage modules are used by ISAR
    # Multiple storage modules can be chosen
    # Each module will be called when storing results from inspections
    # Selecting a different storage module than local may require certain access rights
    STORAGE_LOCAL_ENABLED: bool = Field(default=True)
    STORAGE_BLOB_ENABLED: bool = Field(default=False)
    STORAGE_SLIMM_ENABLED: bool = Field(default=False)

    # Determines whether the MQTT publishing module should be enabled or not
    # The publishing module will attempt to connect to the MQTT broker configured in
    # "service_connections"
    # Options: [false true]
    MQTT_ENABLED: bool = Field(default=False)

    # Determines whether certificate based encryption will be used for the MQTT
    # communication.
    MQTT_SSL_ENABLED: bool = Field(default=True)

    # Determines whether authentication is enabled for the API or not
    # Enabling this requires certain resources available for OAuth2 authentication
    # Currently supported authentication is Azure AD
    # (https://github.com/Intility/fastapi-azure-auth)
    AUTHENTICATION_ENABLED: bool = Field(default=False)

    # Tenant ID for the Azure tenant with your Azure Active Directory application
    AZURE_TENANT_ID: str = Field(default="3aa4a235-b6e2-48d5-9195-7fcf05b459b0")

    # Client ID for ISAR
    AZURE_CLIENT_ID: str = Field(default="fd384acd-5c1b-4c44-a1ac-d41d720ed0fe")
    # If AZURE_CLIENT_ID is set as an environment variable, overwrite this despite missing prefix.
    # This is done to avoid double config of ISAR_AZURE_CLIENT_ID and AZURE_CLIENT_ID.
    # We need the latter as an environment variable for the EnvironmentCredential method for AzureAD.
    azure_client_id_name: str = "AZURE_CLIENT_ID"
    if os.environ.get(azure_client_id_name) is not None:
        print("Using environment variable for AZURE_CLIENT_ID")
        AZURE_CLIENT_ID = os.environ[azure_client_id_name]

    # MQTT username
    # The username and password is set by the MQTT broker and must be known in advance
    # The password should be set as an environment variable "MQTT_PASSWORD"
    # If the password is not set in the environment an empty string will be used
    MQTT_USERNAME: str = Field(default="isar")

    # MQTT host
    MQTT_HOST: str = Field(default="localhost")

    # MQTT port
    MQTT_PORT: int = Field(default=1883)

    # Keyvault name
    KEYVAULT_NAME: str = Field(default="IsarDevKv")

    # Determines whether inspections are uploaded asynchronously or get_inspections in robotinterface
    UPLOAD_INSPECTIONS_ASYNC: bool = Field(default=False)

    # URL to storage account for Azure Blob Storage
    BLOB_STORAGE_ACCOUNT: str = Field(default="eqrobotdevstorage")

    # Name of blob container in Azure Blob Storage [slimm test]
    BLOB_CONTAINER: str = Field(default="test")

    # Client ID for SLIMM App Registration
    SLIMM_CLIENT_ID: str = Field(default="c630ca4d-d8d6-45ab-8cc6-68a363d0de9e")

    # Scope for access to SLIMM Ingestion API
    SLIMM_APP_SCOPE: str = Field(default=".default")

    # URL for SLIMM endpoint
    SLIMM_API_URL: str = Field(
        default="https://scinspectioningestapitest.azurewebsites.net/Ingest"
    )

    # Whether the results should be copied directly into the SLIMM datalake or only the
    # metadata
    COPY_FILES_TO_SLIMM_DATALAKE: bool = Field(default=False)

    # The configuration of this section is tightly coupled with the metadata that is
    # submitted with the results once they have been uploaded.

    # Four digit code indicating facility
    PLANT_CODE: str = Field(default="1320")

    # Name of the facility the robot is operating in
    PLANT_NAME: str = Field(default="Kårstø")

    # Shortname of the facility the robot is operating in
    PLANT_SHORT_NAME: str = Field(default="KAA")

    # Country the robot is operating in
    COUNTRY: str = Field(default="Norway")

    # Type of robot ISAR is monitoring
    ROBOT_TYPE: str = Field(default="robot")

    # Name of robot
    ROBOT_NAME: str = Field(default="Placebot")

    # Unique identifier for this ISAR instance. Note that this should be a generated UUID.
    ISAR_ID: str = Field(default="00000000-0000-0000-0000-000000000000")

    # Serial number of the robot ISAR is connected to
    SERIAL_NUMBER: str = Field(default="0001")

    # Info about robot documentation
    DOCUMENTATION: List[DocumentInfo] = Field(default=[])

    # Data scheme the robot should adhere to
    # Options [DS0001]
    DATA_SCHEME: str = Field(default="DS0001")

    # Coordinate reference system of facility
    COORDINATE_REFERENCE_SYSTEM: str = Field(default="EQUINOR:4100001")

    # Vertical reference system of facility
    VERTICAL_REFERENCE_SYSTEM: str = Field(default="MSL")

    # Rotational representations of reported results
    # Options [quaternion]
    MEDIA_ORIENTATION_REFERENCE_SYSTEM: str = Field(default="quaternion")

    # Contractor who is responsible for robot missions
    CONTRACTOR: str = Field(default="equinor")

    # Timezone
    TIMEZONE: str = Field(default="UTC+00:00")

    # Data classification
    DATA_CLASSIFICATION: str = Field(default="internal")

    # List of MQTT Topics
    TOPIC_ISAR_STATUS: str = Field(default="status", validate_default=True)
    TOPIC_ISAR_MISSION: str = Field(default="mission", validate_default=True)
    TOPIC_ISAR_TASK: str = Field(default="task", validate_default=True)
    TOPIC_ISAR_INSPECTION_RESULT: str = Field(
        default="inspection_result", validate_default=True
    )
    TOPIC_ISAR_ROBOT_INFO: str = Field(default="robot_info", validate_default=True)
    TOPIC_ISAR_ROBOT_HEARTBEAT: str = Field(
        default="robot_heartbeat", validate_default=True
    )

    # Logging

    #   Log handlers
    # Determines which log handlers are used by ISAR
    # Multiple log handlers can be chosen
    # Each handler will be called when logging
    # Selecting a different log handler than local may require certain access rights:
    #    - The Azure AI logger requires the 'APPLICATIONINSIGHTS_CONNECTION_STRING' to be set as an environment variable.
    LOG_HANDLER_LOCAL_ENABLED: bool = Field(default=True)
    LOG_HANDLER_APPLICATION_INSIGHTS_ENABLED: bool = Field(default=False)

    #   Log levels
    API_LOG_LEVEL: str = Field(default="INFO")
    MAIN_LOG_LEVEL: str = Field(default="INFO")
    MQTT_LOG_LEVEL: str = Field(default="INFO")
    STATE_MACHINE_LOG_LEVEL: str = Field(default="INFO")
    UPLOADER_LOG_LEVEL: str = Field(default="INFO")
    CONSOLE_LOG_LEVEL: str = Field(default="INFO")
    URLLIB3_LOG_LEVEL: str = Field(default="WARNING")
    UVICORN_LOG_LEVEL: str = Field(default="WARNING")
    AZURE_LOG_LEVEL: str = Field(default="WARNING")

    LOG_LEVELS: dict = Field(default={})

    REQUIRED_ROLE: str = Field(default="Mission.Control")

    @field_validator("LOG_LEVELS")
    @classmethod
    def set_log_levels(cls, v: Any, info: ValidationInfo) -> dict:
        return {
            "api": info.data["API_LOG_LEVEL"],
            "main": info.data["MAIN_LOG_LEVEL"],
            "mqtt": info.data["MQTT_LOG_LEVEL"],
            "state_machine": info.data["STATE_MACHINE_LOG_LEVEL"],
            "uploader": info.data["UPLOADER_LOG_LEVEL"],
            "console": info.data["CONSOLE_LOG_LEVEL"],
            "urllib3": info.data["URLLIB3_LOG_LEVEL"],
            "uvicorn": info.data["UVICORN_LOG_LEVEL"],
            "azure": info.data["AZURE_LOG_LEVEL"],
        }

    @field_validator(
        "TOPIC_ISAR_STATUS",
        "TOPIC_ISAR_MISSION",
        "TOPIC_ISAR_TASK",
        "TOPIC_ISAR_ROBOT_INFO",
        "TOPIC_ISAR_ROBOT_HEARTBEAT",
        "TOPIC_ISAR_INSPECTION_RESULT",
    )
    @classmethod
    def prefix_isar_topics(cls, v: Any, info: ValidationInfo):
        return f"isar/{info.data['ISAR_ID']}/{v}"

    model_config = SettingsConfigDict(
        env_prefix="ISAR_",
        env_file_encoding="utf-8",
        case_sensitive=True,
        extra="ignore",
    )


load_dotenv()
settings = Settings()


class RobotSettings(BaseSettings):
    def __init__(self) -> None:
        try:
            source = (
                files(f"{settings.ROBOT_PACKAGE}")
                .joinpath("config")
                .joinpath("settings.env")
            )
            with as_file(source) as eml:
                env_file = eml
        except ModuleNotFoundError:
            env_file = None
        super().__init__(_env_file=env_file)

    # ISAR steps the robot is capable of performing
    # This should be set in the robot package settings.env file
    CAPABILITIES: List[str] = Field(default=["return_to_home", "take_image"])

    # Model of the robot which ISAR is connected to
    # This should be set in the robot package settings.env file
    ROBOT_MODEL: RobotModel = Field(default=RobotModel.Robot)  # type: ignore

    # Valid arm poses that the robot may utilize
    # This should be set in the robot package settings.env file
    # Note that if the robot does not support moving an arm this will be None and
    # the functionality will be unavailable
    VALID_ARM_POSES: Optional[List[str]] = Field(default=None)
    model_config = SettingsConfigDict(
        env_file_encoding="utf-8", case_sensitive=True, extra="ignore"
    )


robot_settings = RobotSettings()



import time
from collections import deque
from threading import Thread
from typing import List

import pytest
from injector import Injector
from pytest_mock import MockerFixture

from isar.models.communication.queues.queues import Queues
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine, main
from isar.state_machine.states.idle import Idle
from isar.state_machine.states_enum import States
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorReason,
    RobotException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import TaskStatus
from robot_interface.models.mission.task import ReturnToHome, TakeImage, Task
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from tests.mocks.pose import MockPose
from tests.mocks.robot_interface import (
    MockRobot,
    MockRobotIdleToOfflineToIdleTest,
    MockRobotIdleToBlockedProtectiveStopToIdleTest,
)
from tests.mocks.task import MockTask


class StateMachineThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.state_machine: StateMachine = injector.get(StateMachine)
        self._thread: Thread = Thread(target=main, args=[self.state_machine])
        self._thread.daemon = True

    def start(self):
        self._thread.start()


class UploaderThread(object):
    def __init__(self, injector) -> None:
        self.injector: Injector = injector
        self.uploader: Uploader = Uploader(
            queues=self.injector.get(Queues),
            storage_handlers=injector.get(List[StorageInterface]),
            mqtt_publisher=injector.get(MqttClientInterface),
        )
        self._thread: Thread = Thread(target=self.uploader.run)
        self._thread.daemon = True
        self._thread.start()


@pytest.fixture
def state_machine_thread(injector) -> StateMachineThread:
    return StateMachineThread(injector)


@pytest.fixture
def uploader_thread(injector) -> UploaderThread:
    return UploaderThread(injector=injector)


def test_initial_off(state_machine) -> None:
    assert state_machine.state == "off"


def test_send_status(state_machine) -> None:
    state_machine.send_state_status()
    message = state_machine.queues.state.check()
    assert message == state_machine.current_state


def test_reset_state_machine(state_machine) -> None:
    state_machine.reset_state_machine()

    assert state_machine.current_task is None
    assert state_machine.current_mission is None


def test_state_machine_transitions_when_running_full_mission(
    injector, state_machine_thread
) -> None:
    state_machine_thread.state_machine.run_mission_by_task = False
    state_machine_thread.start()

    task_1: Task = TakeImage(
        target=MockPose.default_pose().position, robot_pose=MockPose.default_pose()
    )
    task_2: Task = ReturnToHome(pose=MockPose.default_pose())
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.21)  # Slightly more than the StateMachine sleep time

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Idle,
        ]
    )


def test_state_machine_failed_dependency(
    injector, state_machine_thread, mocker
) -> None:
    task_1: Task = TakeImage(
        target=MockPose.default_pose().position, robot_pose=MockPose.default_pose()
    )
    task_2: Task = ReturnToHome(pose=MockPose.default_pose())
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    mocker.patch.object(MockRobot, "task_status", return_value=TaskStatus.Failed)

    state_machine_thread.start()

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.21)  # Allow the state machine to transition through the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Idle,
        ]
    )


def test_state_machine_with_successful_collection(
    injector, state_machine_thread, uploader_thread
) -> None:
    state_machine_thread.start()

    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    mission: Mission = Mission(name="Dummy misson", tasks=[MockTask.take_image()])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time

    expected_stored_items = 1
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Idle,
        ]
    )


def test_state_machine_with_unsuccessful_collection(
    injector, mocker, state_machine_thread
) -> None:
    storage_mock: StorageInterface = injector.get(List[StorageInterface])[0]

    mocker.patch.object(MockRobot, "get_inspection", return_value=[])

    state_machine_thread.start()

    mission: Mission = Mission(name="Dummy misson", tasks=[MockTask.take_image()])
    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)

    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time

    expected_stored_items = 0
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Idle,
        ]
    )


def test_state_machine_with_successful_mission_stop(
    injector: Injector,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThread,
) -> None:
    state_machine_thread.start()

    mission: Mission = Mission(name="Dummy misson", tasks=[MockTask.take_image()])
    mocker.patch.object(MockRobot, "task_status", return_value=TaskStatus.InProgress)

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time
    scheduling_utilities.stop_mission()

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Stop,
            States.Idle,
        ]
    )


def test_state_machine_with_unsuccessful_mission_stop(
    injector: Injector,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThread,
    caplog: pytest.LogCaptureFixture,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[MockTask.take_image()])

    scheduling_utilities: SchedulingUtilities = injector.get(SchedulingUtilities)
    mocker.patch.object(MockRobot, "task_status", return_value=TaskStatus.InProgress)
    mocker.patch.object(
        MockRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    state_machine_thread.start()

    scheduling_utilities.start_mission(mission=mission, initial_pose=None)
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time
    scheduling_utilities.stop_mission()

    expected_log = (
        "Be aware that the robot may still be "
        "moving even though a stop has been attempted"
    )
    assert expected_log in caplog.text
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.Idle,
            States.Initialize,
            States.Initiate,
            States.Monitor,
            States.Stop,
            States.Idle,
        ]
    )


def test_state_machine_idle_to_offline_to_idle(mocker, state_machine_thread) -> None:

    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(Idle, "_is_ready_to_poll_for_status", return_value=True)

    state_machine_thread.state_machine.robot = MockRobotIdleToOfflineToIdleTest()
    state_machine_thread.start()
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.Idle, States.Offline, States.Idle]
    )


def test_state_machine_idle_to_blocked_protective_stop_to_idle(
    mocker, state_machine_thread
) -> None:

    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(Idle, "_is_ready_to_poll_for_status", return_value=True)

    state_machine_thread.state_machine.robot = (
        MockRobotIdleToBlockedProtectiveStopToIdleTest()
    )
    state_machine_thread.start()
    time.sleep(0.11)  # Slightly more than the StateMachine sleep time

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.Idle, States.BlockedProtectiveStop, States.Idle]
    )


def _mock_robot_exception_with_message() -> RobotException:
    raise RobotException(
        error_reason=ErrorReason.RobotUnknownErrorException,
        error_description="This is an example error description",
    )



import logging

from injector import inject

from isar.apis.models.models import RobotInfoResponse
from isar.config.settings import robot_settings, settings
from isar.services.utilities.robot_utilities import RobotUtilities


class RobotController:
    @inject
    def __init__(
        self,
        robot_utilities: RobotUtilities,
    ):
        self.robot_utilities: RobotUtilities = robot_utilities
        self.logger = logging.getLogger("api")

    def generate_media_config(self):
        return self.robot_utilities.generate_media_config()

    def get_info(self):
        return RobotInfoResponse(
            robot_package=settings.ROBOT_PACKAGE,
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            robot_capabilities=robot_settings.CAPABILITIES,
            robot_map_name=settings.DEFAULT_MAP,
            plant_short_name=settings.PLANT_SHORT_NAME,
        )



from pathlib import Path
from typing import List

import pytest
from alitra import Frame, Orientation, Pose, Position

from isar.config.settings import settings
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionNotFoundError
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    ReturnToHome,
    TakeImage,
    TakeThermalImage,
)


def test_get_working_mission(mission_reader: LocalPlanner) -> None:
    mission_path = Path("./tests/test_data/test_mission_working.json")
    mission: Mission = mission_reader.read_mission_from_file(mission_path)
    assert isinstance(mission, Mission)


def test_get_mission_with_no_tasks(mission_reader: LocalPlanner) -> None:
    mission_path = Path("./tests/test_data/test_mission_working_no_tasks.json")
    mission: Mission = mission_reader.read_mission_from_file(mission_path)
    assert isinstance(mission, Mission)


def test_read_mission_from_file(mission_reader: LocalPlanner) -> None:
    expected_robot_pose_1 = Pose(
        position=Position(-2, -2, 0, Frame("asset")),
        orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
        frame=Frame("asset"),
    )
    task_1: ReturnToHome = ReturnToHome(pose=expected_robot_pose_1)

    expected_robot_pose_2 = Pose(
        position=Position(-2, 2, 0, Frame("asset")),
        orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
        frame=Frame("asset"),
    )
    expected_inspection_target_1 = Position(2, 2, 0, Frame("asset"))
    task_2: TakeImage = TakeImage(
        target=expected_inspection_target_1, robot_pose=expected_robot_pose_2
    )

    expected_robot_pose_3 = Pose(
        position=Position(2, 2, 0, Frame("asset")),
        orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
        frame=Frame("asset"),
    )
    expected_inspection_target_2 = Position(2, 2, 0, Frame("asset"))
    task_3: TakeImage = TakeImage(
        target=expected_inspection_target_2, robot_pose=expected_robot_pose_3
    )

    expected_robot_pose_4 = Pose(
        position=Position(0, 0, 0, Frame("asset")),
        orientation=Orientation(0, 0, 0.4794255, 0.8775826, Frame("asset")),
        frame=Frame("asset"),
    )
    task_4: ReturnToHome = ReturnToHome(pose=expected_robot_pose_4)

    expected_tasks: List[TASKS] = [
        task_1,
        task_2,
        task_3,
        task_4,
    ]
    mission: Mission = mission_reader.read_mission_from_file(
        Path("./tests/test_data/test_mission_working.json")
    )

    for expected_task, task in zip(expected_tasks, mission.tasks):
        if isinstance(expected_task, ReturnToHome) and isinstance(task, ReturnToHome):
            assert expected_task.pose == task.pose
        if isinstance(expected_task, TakeImage) and isinstance(task, TakeImage):
            assert expected_task.target == task.target
            assert expected_task.robot_pose == task.robot_pose


@pytest.mark.parametrize(
    "mission_path",
    [
        (Path("./tests/test_data/no_file.json")),
        (Path("./tests/test_data/test_mission_not_working.json")),
    ],
)
def test_get_invalid_mission(mission_reader: LocalPlanner, mission_path) -> None:
    with pytest.raises(Exception):
        mission_reader.read_mission_from_file(mission_path)


def test_get_mission_by_id(mission_reader: LocalPlanner) -> None:
    output = mission_reader.get_mission("1")
    assert isinstance(output, Mission)


def test_get_mission_by_invalid_id(mission_reader: LocalPlanner) -> None:
    with pytest.raises(MissionNotFoundError):
        mission_reader.get_mission("12345")


def test_valid_predefined_missions_files(mission_reader: LocalPlanner) -> None:
    # Checks that the predefined mission folder contains only valid missions!
    mission_list_dict = mission_reader.get_predefined_missions()
    predefined_mission_folder = Path(settings.PREDEFINED_MISSIONS_FOLDER)
    assert len(list(predefined_mission_folder.glob("*.json"))) == len(
        list(mission_list_dict)
    )
    for file in predefined_mission_folder.glob("*.json"):
        path_to_file = predefined_mission_folder.joinpath(file.name)
        mission: Mission = mission_reader.read_mission_from_file(path_to_file)
        assert mission is not None


def test_thermal_image_task(mission_reader: LocalPlanner) -> None:
    mission_path: Path = Path("./tests/test_data/test_thermal_image_mission.json")
    output: Mission = mission_reader.read_mission_from_file(mission_path)

    task = output.tasks[0]
    assert isinstance(task, TakeThermalImage)
    assert hasattr(task, "target")
    assert task.type == "take_thermal_image"
    assert hasattr(task, "id")
    assert hasattr(task, "tag_id")



import time
from datetime import datetime
from threading import Thread
from typing import List, Tuple

import pytest
from alitra import Frame, Orientation, Pose, Position

from isar.models.communication.queues.queues import Queues
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.inspection.inspection import ImageMetadata, Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TakeImage
from robot_interface.telemetry.mqtt_client import MqttClientInterface

MISSION_ID = "some-mission-id"
ARBITRARY_IMAGE_METADATA = ImageMetadata(
    start_time=datetime.now(),
    pose=Pose(
        Position(0, 0, 0, Frame("asset")),
        Orientation(x=0, y=0, z=0, w=1, frame=Frame("asset")),
        Frame("asset"),
    ),
    file_type="jpg",
)


@pytest.fixture
def uploader(injector) -> Uploader:
    uploader: Uploader = Uploader(
        queues=injector.get(Queues),
        storage_handlers=injector.get(List[StorageInterface]),
        mqtt_publisher=injector.get(MqttClientInterface),
    )

    # The thread is deliberately started but not joined so that it runs in the
    # background and stops when the test ends
    thread = Thread(target=uploader.run, daemon=True)
    thread.start()

    return uploader


def test_should_upload_from_queue(uploader) -> None:
    take_image_task = TakeImage()
    mission: Mission = Mission(name="Dummy misson", tasks=[take_image_task])

    assert isinstance(mission.tasks[0], TakeImage)
    inspection = Inspection(
        metadata=ARBITRARY_IMAGE_METADATA, id=mission.tasks[0].inspection_id
    )

    message: Tuple[Inspection, Mission] = (
        inspection,
        mission,
    )

    uploader.upload_queue.put(message)
    time.sleep(0.0001)
    assert uploader.storage_handlers[0].blob_exists(inspection)


def test_should_retry_failed_upload_from_queue(uploader) -> None:
    INSPECTION_ID = "123-456"
    inspection = Inspection(metadata=ARBITRARY_IMAGE_METADATA, id=INSPECTION_ID)
    mission: Mission = Mission(name="Dummy Mission")

    message: Tuple[Inspection, Mission] = (
        inspection,
        mission,
    )

    # Need it to fail so that it retries
    uploader.storage_handlers[0].will_fail = True
    uploader.upload_queue.put(message)
    time.sleep(0.0001)

    # Should not upload, instead raise StorageException
    assert not uploader.storage_handlers[0].blob_exists(inspection)
    uploader.storage_handlers[0].will_fail = False
    time.sleep(3)

    # After 3 seconds, it should have retried and now it should be successful
    assert uploader.storage_handlers[0].blob_exists(inspection)



from threading import Lock, Thread
from typing import Any, Optional


class ThreadedRequest:
    def __init__(self, request_func: Any):
        self._thread: Optional[Thread] = None
        self._request_func: Any = request_func
        self._output: Optional[Any] = None
        self._output_lock: Lock = Lock()
        self._exception: Optional[Exception] = None
        self._exception_lock: Lock = Lock()

    def start_thread(self, *request_args, **kwargs) -> bool:
        if self._is_thread_alive():
            return False
        self._output = None
        self._thread = Thread(target=self._thread_func, args=request_args, **kwargs)
        self._thread.start()
        return True

    def get_output(self) -> Any:
        if self._is_thread_alive():
            raise ThreadedRequestNotFinishedError

        self._exception_lock.acquire()
        exception = self._exception
        self._exception_lock.release()

        if exception:
            raise exception

        self._output_lock.acquire()
        output = self._output
        self._output_lock.release()

        return output

    def wait_for_thread(self) -> None:
        if not self._thread:
            return
        self._thread.join()

    def _is_thread_alive(self) -> bool:
        if not self._thread:
            return False
        return self._thread.is_alive()

    def _thread_func(self, *args) -> None:
        try:
            request_output: Any = self._request_func(*args)
        except Exception as e:
            self._exception_lock.acquire()
            self._exception = e
            self._exception_lock.release()
            return

        self._output_lock.acquire()
        self._output = request_output
        self._output_lock.release()


class ThreadedRequestError(Exception):
    pass


class ThreadedRequestNotFinishedError(ThreadedRequestError):
    pass



import logging
import time
from typing import TYPE_CHECKING, Any, Callable, Optional

from transitions import State

from isar.config.settings import settings
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotInfeasibleMissionException,
)

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Initiate(State):
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="initiate", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.initiate_failure_counter: int = 0
        self.initiate_failure_counter_limit: int = (
            settings.INITIATE_FAILURE_COUNTER_LIMIT
        )
        self.logger = logging.getLogger("state_machine")

        self.initiate_thread: Optional[ThreadedRequest] = None

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        self.initiate_failure_counter = 0
        if self.initiate_thread:
            self.initiate_thread.wait_for_thread()
        self.initiate_thread = None

    def _run(self) -> None:
        transition: Callable
        while True:
            if self.state_machine.should_stop_mission():
                transition = self.state_machine.stop  # type: ignore
                break

            if not self.initiate_thread:
                self._run_initiate_thread(
                    initiate_function=self.state_machine.robot.initiate_mission,
                    function_argument=self.state_machine.current_mission,
                    thread_name="State Machine Initiate Mission",
                )

            try:
                self.initiate_thread.get_output()
                transition = self.state_machine.initiated  # type: ignore
                break
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue

            except RobotInfeasibleMissionException as e:
                self.state_machine.current_mission.error_message = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                self.logger.warning(
                    f"Failed to initiate mission "
                    f"{str(self.state_machine.current_mission.id)[:8]} because: "
                    f"{e.error_description}"
                )
                transition = self.state_machine.initiate_failed  # type: ignore
                break

            except RobotException as e:
                self.initiate_thread = None
                self.initiate_failure_counter += 1
                self.logger.warning(
                    f"Initiating failed #: {str(self.initiate_failure_counter)} "
                    f"because: {e.error_description}"
                )

                if self.initiate_failure_counter >= self.initiate_failure_counter_limit:
                    self.state_machine.current_task.error_message = ErrorMessage(
                        error_reason=e.error_reason,
                        error_description=e.error_description,
                    )
                    self.logger.error(
                        f"Mission will be cancelled after failing to initiate "
                        f"{self.initiate_failure_counter_limit} times because: "
                        f"{e.error_description}"
                    )
                    transition = self.state_machine.initiate_failed  # type: ignore
                    break

            time.sleep(self.state_machine.sleep_time)

        transition()

    def _run_initiate_thread(
        self, initiate_function: Callable, function_argument: Any, thread_name: str
    ) -> None:
        self.initiate_thread = ThreadedRequest(request_func=initiate_function)

        self.initiate_thread.start_thread(
            function_argument,
            name=thread_name,
        )



from dataclasses import dataclass
from enum import Enum


class MediaConnectionType(str, Enum):
    LiveKit = "LiveKit"


@dataclass
class MediaConfig:
    url: str
    token: str
    media_connection_type: MediaConnectionType



import logging
import time
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Optional, Tuple, Union

from injector import inject
from transitions import State

from isar.config.settings import settings
from isar.services.utilities.threaded_request import (
    ThreadedRequest,
    ThreadedRequestNotFinishedError,
)
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotCommunicationException,
    RobotCommunicationTimeoutException,
    RobotException,
    RobotRetrieveInspectionException,
    RobotTaskStatusException,
)
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import TaskStatus
from robot_interface.models.mission.task import InspectionTask, Task

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(State):
    @inject
    def __init__(self, state_machine: "StateMachine") -> None:
        super().__init__(name="monitor", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.request_status_failure_counter: int = 0
        self.request_status_failure_counter_limit: int = (
            settings.REQUEST_STATUS_FAILURE_COUNTER_LIMIT
        )

        self.logger = logging.getLogger("state_machine")
        self.task_status_thread: Optional[ThreadedRequest] = None

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        if self.task_status_thread:
            self.task_status_thread.wait_for_thread()
        self.task_status_thread = None

    def _run(self) -> None:
        transition: Callable
        while True:
            if self.state_machine.should_stop_mission():
                transition = self.state_machine.stop  # type: ignore
                break

            if self.state_machine.should_pause_mission():
                transition = self.state_machine.pause  # type: ignore
                break

            if not self.task_status_thread:
                self._run_get_status_thread(
                    status_function=self.state_machine.robot.task_status,
                    function_argument=self.state_machine.current_task.id,
                    thread_name="State Machine Monitor Get Task Status",
                )
            try:
                status: TaskStatus = self.task_status_thread.get_output()
            except ThreadedRequestNotFinishedError:
                time.sleep(self.state_machine.sleep_time)
                continue
            except (
                RobotCommunicationTimeoutException,
                RobotCommunicationException,
            ) as e:
                retry_limit_exceeded: bool = self._check_if_exceeded_retry_limit(e)
                if retry_limit_exceeded:
                    self.logger.error(
                        f"Monitoring task {self.state_machine.current_task.id[:8]} failed "
                        f"because: {e.error_description}"
                    )
                    status = TaskStatus.Failed
                else:
                    time.sleep(settings.REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY)
                    continue

            except RobotTaskStatusException as e:
                self.state_machine.current_task.error_message = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                self.logger.error(
                    f"Monitoring task {self.state_machine.current_task.id[:8]} failed "
                    f"because: {e.error_description}"
                )
                status = TaskStatus.Failed

            except RobotException as e:
                self._set_error_message(e)
                status = TaskStatus.Failed

                self.logger.error(
                    f"Retrieving the status failed because: {e.error_description}"
                )

            if not isinstance(status, TaskStatus):
                self.logger.error(
                    f"Received an invalid status update {status} when monitoring mission. "
                    "Only TaskStatus is expected."
                )
                break

            if self.state_machine.current_task is None:
                self.state_machine.iterate_current_task()

            self.state_machine.current_task.status = status

            if (
                not settings.UPLOAD_INSPECTIONS_ASYNC
                and self._should_upload_inspections()
            ):
                get_inspection_thread = ThreadedRequest(
                    self._queue_inspections_for_upload
                )
                get_inspection_thread.start_thread(
                    deepcopy(self.state_machine.current_mission),
                    deepcopy(self.state_machine.current_task),
                    name="State Machine Get Inspections",
                )

            if self.state_machine.current_task.is_finished():
                self._report_task_status(self.state_machine.current_task)
                self.state_machine.publish_task_status(
                    task=self.state_machine.current_task
                )

                self.state_machine.iterate_current_task()
                if self.state_machine.current_task is None:
                    transition = self.state_machine.mission_finished  # type: ignore
                    break

                # Report and update next task
                self.state_machine.current_task.update_task_status()
                self.state_machine.publish_task_status(
                    task=self.state_machine.current_task
                )

            self.task_status_thread = None
            time.sleep(self.state_machine.sleep_time)

        transition()

    def _run_get_status_thread(
        self, status_function: Callable, function_argument: str, thread_name: str
    ) -> None:
        self.task_status_thread = ThreadedRequest(request_func=status_function)
        self.task_status_thread.start_thread(function_argument, name=thread_name)

    def _queue_inspections_for_upload(
        self, mission: Mission, current_task: InspectionTask
    ) -> None:
        try:
            inspection: Inspection = self.state_machine.robot.get_inspection(
                task=current_task
            )
            if current_task.inspection_id != inspection.id:
                self.logger.warning(
                    f"The inspection_id of task ({current_task.inspection_id}) "
                    f"and result ({inspection.id}) is not matching. "
                    f"This may lead to confusions when accessing the inspection later"
                )

        except (RobotRetrieveInspectionException, RobotException) as e:
            self._set_error_message(e)
            self.logger.error(
                f"Failed to retrieve inspections because: {e.error_description}"
            )
            return

        if not inspection:
            self.logger.warning(
                f"No inspection result data retrieved for task {str(current_task.id)[:8]}"
            )

        inspection.metadata.tag_id = current_task.tag_id

        message: Tuple[Inspection, Mission] = (
            inspection,
            mission,
        )
        self.state_machine.queues.upload_queue.put(message)
        self.logger.info(
            f"Inspection result: {str(inspection.id)[:8]} queued for upload"
        )

    def _report_task_status(self, task: Task) -> None:
        self.request_status_failure_counter = 0
        if task.status == TaskStatus.Failed:
            self.logger.warning(
                f"Task: {str(task.id)[:8]} was reported as failed by the robot"
            )
        elif task.status == TaskStatus.Successful:
            self.logger.info(
                f"{type(task).__name__} task: {str(task.id)[:8]} completed"
            )

    def _should_upload_inspections(self) -> bool:
        return (
            self.state_machine.current_task.is_finished()
            and self.state_machine.current_task.status == TaskStatus.Successful
            and isinstance(self.state_machine.current_task, InspectionTask)
        )

    def _set_error_message(self, e: RobotException) -> None:
        error_message: ErrorMessage = ErrorMessage(
            error_reason=e.error_reason, error_description=e.error_description
        )
        self.state_machine.current_task.error_message = error_message

    def _check_if_exceeded_retry_limit(
        self, e: Union[RobotCommunicationTimeoutException, RobotCommunicationException]
    ) -> bool:
        self.state_machine.current_mission.error_message = ErrorMessage(
            error_reason=e.error_reason, error_description=e.error_description
        )
        self.task_status_thread = None
        self.request_status_failure_counter += 1
        self.logger.warning(
            f"Monitoring task {self.state_machine.current_task.id} failed #: "
            f"{self.request_status_failure_counter} failed because: {e.error_description}"
        )

        if (
            self.request_status_failure_counter
            >= self.request_status_failure_counter_limit
        ):
            self.state_machine.current_task.error_message = ErrorMessage(
                error_reason=e.error_reason,
                error_description=e.error_description,
            )
            self.logger.error(
                f"Task will be cancelled after failing to get task status "
                f"{self.request_status_failure_counter} times because: "
                f"{e.error_description}"
            )
            return True

        return False



from isar.apis.models.models import (
    InputOrientation,
    InputPose,
    InputPosition,
    StartMissionResponse,
    TaskResponse,
)
from isar.apis.models.start_mission_definition import (
    InspectionTypes,
    StartMissionDefinition,
    StartMissionInspectionDefinition,
    StartMissionTaskDefinition,
)
from robot_interface.models.mission.mission import Mission
from tests.mocks.task import MockTask


class MockMissionDefinition:
    mock_input_position = InputPosition(x=1, y=1, z=1, frame_name="robot")
    mock_input_orientation = InputOrientation(x=0, y=0, z=0, w=0, frame_name="robot")
    mock_input_pose = InputPose(
        position=mock_input_position,
        orientation=mock_input_orientation,
        frame_name="robot",
    )
    mock_input_target_position = InputPosition(x=5, y=5, z=5, frame_name="robot")
    mock_task_take_image = MockTask.take_image()
    mock_task_return_home = MockTask.return_home()
    default_mission = Mission(
        id="default_mission",
        name="Dummy misson",
        tasks=[
            mock_task_take_image,
            mock_task_return_home,
        ],
    )
    mock_start_mission_inspection_definition = StartMissionInspectionDefinition(
        type=InspectionTypes.image,
        inspection_target=mock_input_target_position,
    )
    mock_task_response_take_image = TaskResponse(
        id=mock_task_take_image.id,
        tag_id=mock_task_take_image.tag_id,
        inspection_id=mock_task_take_image.inspection_id,
        type=mock_task_take_image.type,
    )

    mock_task_response_return_home = TaskResponse(
        id=mock_task_return_home.id,
        tag_id=mock_task_return_home.tag_id,
        type=mock_task_return_home.type,
    )
    mock_start_mission_response = StartMissionResponse(
        id=default_mission.id,
        tasks=[mock_task_response_take_image, mock_task_response_return_home],
    )
    mock_start_mission_definition = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspection=mock_start_mission_inspection_definition,
            ),
        ]
    )
    mock_start_mission_definition_task_ids = StartMissionDefinition(
        tasks=[
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspection=mock_start_mission_inspection_definition,
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspection=mock_start_mission_inspection_definition,
            ),
            StartMissionTaskDefinition(
                pose=mock_input_pose,
                tag="dummy_tag",
                inspection=mock_start_mission_inspection_definition,
            ),
        ]
    )



import json
import logging

from azure.identity import DefaultAzureCredential
from injector import inject
from requests import HTTPError, RequestException
from requests_toolbelt import MultipartEncoder

from isar.config.settings import settings
from isar.services.auth.azure_credentials import AzureCredentials
from isar.services.service_connections.request_handler import RequestHandler
from isar.storage.storage_interface import StorageException, StorageInterface
from isar.storage.utilities import get_filename
from robot_interface.models.inspection.inspection import Inspection, ThermalVideo, Video
from robot_interface.models.mission.mission import Mission


class SlimmStorage(StorageInterface):
    @inject
    def __init__(self, request_handler: RequestHandler) -> None:
        self.request_handler: RequestHandler = request_handler
        self.logger = logging.getLogger("uploader")

        self.credentials: DefaultAzureCredential = (
            AzureCredentials.get_azure_credentials()
        )

        client_id: str = settings.SLIMM_CLIENT_ID
        scope: str = settings.SLIMM_APP_SCOPE
        self.request_scope: str = f"{client_id}/{scope}"

        self.url: str = settings.SLIMM_API_URL

    def store(self, inspection: Inspection, mission: Mission) -> str:
        filename: str = get_filename(
            inspection=inspection,
        )
        filename = f"{filename}.{inspection.metadata.file_type}"
        if type(inspection) in [Video, ThermalVideo]:
            inspection_path = self._store_video(filename, inspection, mission)
        else:
            inspection_path = self._store_image(filename, inspection, mission)
        return inspection_path

    def _store_image(
        self, filename: str, inspection: Inspection, mission: Mission
    ) -> str:
        multiform_body: MultipartEncoder = self._construct_multiform_request_image(
            filename=filename, inspection=inspection, mission=mission
        )
        request_url: str = f"{self.url}/UploadSingleImage"
        inspection_path = self._ingest(
            inspection=inspection,
            multiform_body=multiform_body,
            request_url=request_url,
        )
        return inspection_path

    def _store_video(
        self, filename: str, inspection: Inspection, mission: Mission
    ) -> str:
        multiform_body: MultipartEncoder = self._construct_multiform_request_video(
            filename=filename, inspection=inspection, mission=mission
        )
        request_url = f"{self.url}/UploadSingleVideo"
        inspection_path = self._ingest(
            inspection=inspection,
            multiform_body=multiform_body,
            request_url=request_url,
        )
        return inspection_path

    def _ingest(
        self,
        inspection: Inspection,
        multiform_body: MultipartEncoder,
        request_url: str,
    ) -> str:
        token: str = self.credentials.get_token(self.request_scope).token
        try:
            response = self.request_handler.post(
                url=request_url,
                data=multiform_body,
                headers={
                    "Authorization": f"Bearer {token}",
                    "Content-Type": multiform_body.content_type,
                },
            )
            guid = json.loads(response.text)["guid"]
            self.logger.info(f"SLIMM upload GUID: {guid}")
        except (RequestException, HTTPError) as e:
            self.logger.warning(
                f"Failed to upload inspection: {inspection.id} to SLIMM due to a "
                f"request exception"
            )
            raise StorageException from e
        data = json.loads(response.content)
        return data["guid"]

    @staticmethod
    def _construct_multiform_request_image(
        filename: str, inspection: Inspection, mission: Mission
    ):
        array_of_orientation = (
            inspection.metadata.pose.orientation.to_quat_array().tolist()
        )
        multiform_body: MultipartEncoder = MultipartEncoder(
            fields={
                "PlantFacilitySAPCode": settings.PLANT_CODE,
                "InstCode": settings.PLANT_SHORT_NAME,
                "InternalClassification": settings.DATA_CLASSIFICATION,
                "IsoCountryCode": "NO",
                "Geodetic.CoordinateReferenceSystemCode": settings.COORDINATE_REFERENCE_SYSTEM,  # noqa: E501
                "Geodetic.VerticalCoordinateReferenceSystemCode": settings.VERTICAL_REFERENCE_SYSTEM,  # noqa: E501
                "Geodetic.OrientationReferenceSystem": settings.MEDIA_ORIENTATION_REFERENCE_SYSTEM,  # noqa: E501
                "SensorCarrier.SensorCarrierId": settings.ISAR_ID,
                "SensorCarrier.ModelName": settings.ROBOT_TYPE,
                "Mission.MissionId": mission.id,
                "Mission.Client": "Equinor",
                "ImageMetadata.Timestamp": inspection.metadata.start_time.isoformat(),  # noqa: E501
                "ImageMetadata.X": str(inspection.metadata.pose.position.x),
                "ImageMetadata.Y": str(inspection.metadata.pose.position.y),
                "ImageMetadata.Z": str(inspection.metadata.pose.position.z),
                "ImageMetadata.CameraOrientation1": str(array_of_orientation[0]),
                "ImageMetadata.CameraOrientation2": str(array_of_orientation[1]),
                "ImageMetadata.CameraOrientation3": str(array_of_orientation[2]),
                "ImageMetadata.CameraOrientation4": str(array_of_orientation[3]),
                "ImageMetadata.AnalysisMethods": (
                    inspection.metadata.additional["analysis_type"]
                    if inspection.metadata.additional
                    and inspection.metadata.additional["analysis_type"]
                    else "N/A"
                ),
                "ImageMetadata.Description": str(inspection.metadata.additional),
                "ImageMetadata.FunctionalLocation": (
                    inspection.metadata.tag_id  # noqa: E501
                    if inspection.metadata.tag_id
                    else "N/A"
                ),
                "Filename": filename,
                "AttachedFile": (filename, inspection.data),
            }
        )
        return multiform_body

    @staticmethod
    def _construct_multiform_request_video(
        filename: str,
        inspection: Inspection,
        mission: Mission,
    ):
        array_of_orientation = (
            inspection.metadata.pose.orientation.to_quat_array().tolist()
        )
        multiform_body: MultipartEncoder = MultipartEncoder(
            fields={
                "PlantFacilitySAPCode": settings.PLANT_CODE,
                "InstCode": settings.PLANT_SHORT_NAME,
                "InternalClassification": settings.DATA_CLASSIFICATION,
                "IsoCountryCode": "NO",
                "Geodetic.CoordinateReferenceSystemCode": settings.COORDINATE_REFERENCE_SYSTEM,  # noqa: E501
                "Geodetic.VerticalCoordinateReferenceSystemCode": settings.VERTICAL_REFERENCE_SYSTEM,  # noqa: E501
                "Geodetic.OrientationReferenceSystem": settings.MEDIA_ORIENTATION_REFERENCE_SYSTEM,  # noqa: E501
                "SensorCarrier.SensorCarrierId": settings.ISAR_ID,
                "SensorCarrier.ModelName": settings.ROBOT_TYPE,
                "Mission.MissionId": mission.id,
                "Mission.Client": "Equinor",
                "VideoMetadata.Timestamp": inspection.metadata.start_time.isoformat(),  # noqa: E501
                # Converting to int because SLIMM expects an int, while we use floats in operations.
                "VideoMetadata.Duration": str(int(inspection.metadata.duration)),  # type: ignore
                "VideoMetadata.X": str(inspection.metadata.pose.position.x),
                "VideoMetadata.Y": str(inspection.metadata.pose.position.y),
                "VideoMetadata.Z": str(inspection.metadata.pose.position.z),
                "VideoMetadata.CameraOrientation1": str(array_of_orientation[0]),
                "VideoMetadata.CameraOrientation2": str(array_of_orientation[1]),
                "VideoMetadata.CameraOrientation3": str(array_of_orientation[2]),
                "VideoMetadata.CameraOrientation4": str(array_of_orientation[3]),
                "VideoMetadata.AnalysisMethods": (
                    inspection.metadata.additional["analysis_type"]
                    if inspection.metadata.additional
                    and inspection.metadata.additional["analysis_type"]
                    else "N/A"
                ),
                "VideoMetadata.Description": str(inspection.metadata.additional),
                "VideoMetadata.FunctionalLocation": (
                    inspection.metadata.tag_id  # noqa: E501
                    if inspection.metadata.tag_id
                    else "N/A"
                ),
                "Filename": filename,
                "AttachedFile": (filename, inspection.data),
            }
        )
        return multiform_body



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Union

from alitra import Pose

from robot_interface.models.robots.battery_state import BatteryState
from robot_interface.models.mission.status import RobotStatus, MissionStatus, TaskStatus
from robot_interface.models.mission.task import TaskTypes
from robot_interface.models.exceptions.robot_exceptions import ErrorReason


@dataclass
class TelemetryPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime


@dataclass
class CloudHealthPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime


@dataclass
class TelemetryPosePayload(TelemetryPayload):
    pose: Pose


@dataclass
class TelemetryBatteryPayload(TelemetryPayload):
    battery_level: float
    battery_state: Optional[BatteryState] = None


@dataclass
class TelemetryObstacleStatusPayload(TelemetryPayload):
    obstacle_status: bool


@dataclass
class TelemetryPressurePayload(TelemetryPayload):
    pressure_level: float


@dataclass
class DocumentInfo:
    name: str
    url: str


@dataclass
class RobotStatusPayload:
    isar_id: str
    robot_name: str
    status: RobotStatus
    timestamp: datetime


@dataclass
class RobotInfoPayload:
    isar_id: str
    robot_name: str
    robot_model: str
    robot_serial_number: str
    robot_asset: str
    documentation: List[DocumentInfo]
    host: str
    port: int
    capabilities: List[str]
    timestamp: datetime


@dataclass
class RobotHeartbeatPayload:
    isar_id: str
    robot_name: str
    timestamp: datetime


@dataclass
class MissionPayload:
    isar_id: str
    robot_name: str
    mission_id: Optional[str]
    status: Optional[MissionStatus]
    error_reason: Optional[ErrorReason]
    error_description: Optional[str]
    timestamp: datetime


@dataclass
class TaskPayload:
    isar_id: str
    robot_name: str
    mission_id: Optional[str]
    task_id: Optional[str]
    status: Optional[TaskStatus]
    task_type: Optional[TaskTypes]
    error_reason: Optional[ErrorReason]
    error_description: Optional[str]
    timestamp: datetime


@dataclass
class InspectionResultPayload:
    isar_id: str
    robot_name: str
    inspection_id: str
    inspection_path: Union[str, dict]
    installation_code: str
    analysis_to_be_run: Optional[str]
    timestamp: datetime



import json
import re
from http import HTTPStatus
from unittest import mock

import pytest
from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from isar.apis.models.models import ControlMissionResponse
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from tests.mocks.mission_definition import MockMissionDefinition

mock_mission = MockMissionDefinition.default_mission

mock_return_off = mock.Mock(return_value=States.Off)
mock_return_idle = mock.Mock(return_value=States.Idle)
mock_return_monitor = mock.Mock(return_value=States.Monitor)
mock_return_paused = mock.Mock(return_value=States.Paused)
mock_void = mock.Mock()

mock_task = mock_mission.tasks[0]
mock_start_mission_response = MockMissionDefinition.mock_start_mission_response
mock_control_mission_response = ControlMissionResponse(
    mission_id=mock_mission.id,
    mission_status=mock_mission.status,
    task_id=mock_task.id,
    task_status=mock_task.status,
)
mock_return_control_mission_response = mock.Mock(
    return_value=mock_control_mission_response
)
mock_queue_timeout_error = mock.Mock(side_effect=QueueTimeoutError)
mock_mission_planner_error = mock.Mock(side_effect=MissionPlannerError)


class TestStartMissionByID:
    schedule_start_mission_path = "/schedule/start-mission"
    mock_get_mission = mock.Mock(return_value=mock_mission)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "get_mission", mock_get_mission)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_start_mission_by_id(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(mock_start_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "get_mission", mock_get_mission)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_state_machine_in_conflicting_state(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_mission_not_found(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/9999")
        assert response.status_code == HTTPStatus.NOT_FOUND
        assert response.json() == {"detail": "Mission with id '9999' not found"}

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "get_mission", mock_get_mission)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_start_mission_timeout(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to start mission in ISAR"
        }

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "get_mission", mock_get_mission)
    @mock.patch("isar.config.settings.robot_settings.CAPABILITIES", [])
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_robot_not_capable(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.BAD_REQUEST
        response_detail = response.json()["detail"]
        assert re.match(
            "Bad Request - Robot is not capable of performing mission.", response_detail
        )
        assert re.search("return_to_home", response_detail)
        assert re.search("take_image", response_detail)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    @mock.patch.object(LocalPlanner, "get_mission", mock_mission_planner_error)
    def test_mission_planner_error(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {"detail": "Could not plan mission"}


class TestStartMission:
    schedule_start_mission_path = "/schedule/start-mission"
    mock_start_mission_definition = MockMissionDefinition.mock_start_mission_definition
    mock_start_mission_content = {"mission_definition": mock_start_mission_definition}

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_start_mission(self, client: TestClient):
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.mock_start_mission_content),
        )
        assert response.status_code == HTTPStatus.OK

    def test_incomplete_request(self, client: TestClient):
        response = client.post(url=self.schedule_start_mission_path, json={})
        assert response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    def test_state_machine_in_conflicting_state(self, client: TestClient):
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.mock_start_mission_content),
        )
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_start_mission_timeout(self, client: TestClient):
        response = client.post(
            url=self.schedule_start_mission_path,
            json=jsonable_encoder(self.mock_start_mission_content),
        )
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to start mission in ISAR"
        }

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch("isar.config.settings.robot_settings.CAPABILITIES", [])
    @mock.patch.object(SchedulingUtilities, "start_mission", mock_void)
    def test_robot_not_capable(self, client: TestClient):
        response = client.post(url=f"{self.schedule_start_mission_path}/1")
        assert response.status_code == HTTPStatus.BAD_REQUEST
        response_detail = response.json()["detail"]
        assert re.match(
            "Bad Request - Robot is not capable of performing mission.", response_detail
        )
        assert re.search("return_to_home", response_detail)
        assert re.search("take_image", response_detail)


class TestPauseMission:
    schedule_pause_mission_path = "/schedule/pause-mission"

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(
        SchedulingUtilities, "_send_command", mock_return_control_mission_response
    )
    def test_pause_mission(self, client: TestClient):
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(mock_control_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    def test_state_machine_in_conflicting_state(self, client: TestClient):
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_pause_mission_timeout(self, client: TestClient):
        response = client.post(url=self.schedule_pause_mission_path)
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to pause mission"
        }


class TestResumeMission:
    schedule_resume_mission_path = "/schedule/resume-mission"

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_paused)
    @mock.patch.object(
        SchedulingUtilities, "_send_command", mock_return_control_mission_response
    )
    def test_resume_mission(self, client: TestClient):
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(mock_control_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    def test_state_machine_in_conflicting_state(self, client: TestClient):
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_paused)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_resume_mission_timeout(self, client: TestClient):
        response = client.post(url=self.schedule_resume_mission_path)
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to resume mission"
        }


class TestStopMission:
    schedule_stop_mission_path = "/schedule/stop-mission"
    valid_states = [
        States.Initiate,
        States.Initialize,
        States.Idle,
        States.Monitor,
        States.Paused,
        States.Stop,
    ]

    @pytest.mark.parametrize("state", valid_states)
    @mock.patch.object(
        SchedulingUtilities, "_send_command", mock_return_control_mission_response
    )
    def test_stop_mission(
        self, client: TestClient, state: States, mocker: MockerFixture
    ):
        mocker.patch.object(SchedulingUtilities, "get_state", return_value=state)
        response = client.post(url=self.schedule_stop_mission_path)
        assert response.status_code == HTTPStatus.OK
        assert response.json() == jsonable_encoder(mock_control_mission_response)

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_off)
    @mock.patch.object(
        SchedulingUtilities, "stop_mission", mock_control_mission_response
    )
    def test_can_not_stop_mission_in_off(self, client: TestClient):
        response = client.post(url=self.schedule_stop_mission_path)
        assert response.status_code == HTTPStatus.CONFLICT

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_stop_mission_timeout(self, client: TestClient):
        response = client.post(url=self.schedule_stop_mission_path)
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR


class TestDriveTo:
    schedule_drive_to_path = "/schedule/drive-to"
    mock_target_pose = MockMissionDefinition.mock_input_pose
    mock_data: str = json.dumps(jsonable_encoder(mock_target_pose))

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_void)
    def test_drive_to(self, client: TestClient):
        response = client.post(url=self.schedule_drive_to_path, data=self.mock_data)
        assert response.status_code == HTTPStatus.OK

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_idle)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_queue_timeout_error)
    def test_drive_to_timeout(self, client: TestClient):
        response = client.post(url=self.schedule_drive_to_path, data=self.mock_data)
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.json() == {
            "detail": "Internal Server Error - Failed to start mission in ISAR"
        }

    @mock.patch.object(SchedulingUtilities, "get_state", mock_return_monitor)
    @mock.patch.object(SchedulingUtilities, "_send_command", mock_void)
    def test_state_machine_in_conflicting_state(self, client: TestClient):
        response = client.post(url=self.schedule_drive_to_path, data=self.mock_data)
        assert response.status_code == HTTPStatus.CONFLICT


class TestInfoRobotSettings:
    def test_info_robot_settings(self, client: TestClient):
        response = client.get(url="/info/robot-settings")
        assert response.status_code == HTTPStatus.OK



from dataclasses import dataclass
from typing import Optional

from alitra import Pose

from robot_interface.models.mission.mission import Mission


@dataclass
class StartMissionMessage:
    mission: Mission
    initial_pose: Optional[Pose]



import json
import time
from datetime import datetime, timezone
from queue import Queue

from isar.config.settings import robot_settings, settings
from robot_interface.telemetry.mqtt_client import MqttPublisher
from robot_interface.telemetry.payloads import RobotInfoPayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class RobotInfoPublisher:
    def __init__(self, mqtt_queue: Queue):
        self.mqtt_publisher: MqttPublisher = MqttPublisher(mqtt_queue=mqtt_queue)

    def run(self) -> None:
        while True:
            payload: RobotInfoPayload = RobotInfoPayload(
                isar_id=settings.ISAR_ID,
                robot_name=settings.ROBOT_NAME,
                robot_model=robot_settings.ROBOT_MODEL,  # type: ignore
                robot_serial_number=settings.SERIAL_NUMBER,
                robot_asset=settings.PLANT_SHORT_NAME,
                documentation=settings.DOCUMENTATION,
                host=settings.API_HOST_VIEWED_EXTERNALLY,
                port=settings.API_PORT,
                capabilities=robot_settings.CAPABILITIES,
                timestamp=datetime.now(timezone.utc),
            )

            self.mqtt_publisher.publish(
                topic=settings.TOPIC_ISAR_ROBOT_INFO,
                payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            )

            time.sleep(settings.ROBOT_INFO_PUBLISH_INTERVAL)



from collections import deque
from queue import Empty, Queue
from typing import Any


class StatusQueue(Queue):
    def __init__(self) -> None:
        super().__init__()

    def check(self) -> Any:
        if not self._qsize():
            raise Empty
        with self.mutex:
            queueList = list(self.queue)
            return queueList.pop()

    def update(self, item: Any):
        with self.mutex:
            self.queue = deque()
            self.queue.append(item)



from datetime import datetime
from queue import Queue
from threading import Thread
from typing import Callable, List

from alitra import Frame, Orientation, Pose, Position

from robot_interface.models.initialize.initialize_params import InitializeParams
from robot_interface.models.inspection.inspection import (
    Image,
    ImageMetadata,
    Inspection,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import InspectionTask, Task
from robot_interface.models.robots.media import MediaConfig, MediaConnectionType
from robot_interface.robot_interface import RobotInterface


class MockRobot(RobotInterface):
    def __init__(
        self,
        mission_status: MissionStatus = MissionStatus.Successful,
        task_status: TaskStatus = TaskStatus.Successful,
        stop: bool = True,
        pose: Pose = Pose(
            position=Position(x=0, y=0, z=0, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        ),
        robot_status: RobotStatus = RobotStatus.Available,
    ):
        self.mission_status_return_value: MissionStatus = mission_status
        self.task_status_return_value: TaskStatus = task_status
        self.stop_return_value: bool = stop
        self.robot_pose_return_value: Pose = pose
        self.robot_status_return_value: RobotStatus = robot_status

    def initiate_mission(self, mission: Mission) -> None:
        return

    def initiate_task(self, task: Task) -> None:
        return

    def task_status(self, task_id: str) -> TaskStatus:
        return self.task_status_return_value

    def stop(self) -> None:
        return

    def pause(self) -> None:
        return

    def resume(self) -> None:
        return

    def get_inspection(self, task: InspectionTask) -> Inspection:
        return Image(
            metadata=mock_image_metadata(),
            id=task.inspection_id,
            data=b"Some binary image data",
        )

    def generate_media_config(self) -> MediaConfig:
        return MediaConfig(
            url="mockURL",
            token="mockToken",
            media_connection_type=MediaConnectionType.LiveKit,
        )

    def register_inspection_callback(
        self, callback_function: Callable[[Inspection, Mission], None]
    ) -> None:
        return

    def initialize(self, params: InitializeParams) -> None:
        return

    def get_telemetry_publishers(
        self, queue: Queue, isar_id: str, robot_name: str
    ) -> List[Thread]:
        return []

    def robot_status(self) -> RobotStatus:
        return self.robot_status_return_value


def mock_image_metadata() -> ImageMetadata:
    return ImageMetadata(
        start_time=datetime.now(),
        pose=Pose(
            Position(0, 0, 0, Frame("robot")),
            Orientation(0, 0, 0, 1, Frame("robot")),
            Frame("robot"),
        ),
        file_type="jpg",
    )


class MockRobotIdleToOfflineToIdleTest(MockRobot):
    def __init__(self):
        self.first = True

    def robot_status(self) -> RobotStatus:
        if self.first:
            self.first = False
            return RobotStatus.Offline

        return RobotStatus.Available


class MockRobotIdleToBlockedProtectiveStopToIdleTest(MockRobot):
    def __init__(self):
        self.first = True

    def robot_status(self) -> RobotStatus:
        if self.first:
            self.first = False
            return RobotStatus.BlockedProtectiveStop

        return RobotStatus.Available



import logging
from copy import deepcopy
from http import HTTPStatus
from queue import Empty
from typing import Any, List, Optional, Set

from alitra import Pose
from fastapi import HTTPException
from injector import inject
from requests import HTTPError

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import (
    MissionNotFoundError,
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues.queue_io import QueueIO
from isar.models.communication.queues.queues import Queues
from isar.models.communication.queues.queue_timeout_error import (
    QueueTimeoutError,
)
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    @inject
    def __init__(
        self,
        queues: Queues,
        mission_planner: MissionPlannerInterface,
        queue_timeout: int = settings.QUEUE_TIMEOUT,
    ):
        self.queues: Queues = queues
        self.mission_planner: MissionPlannerInterface = mission_planner
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        """Return the current state of the state machine

        Raises
        ------
        HTTPException 500 Internal Server Error
            If the current state is not available on the queue
        """
        try:
            return self.queues.state.check()
        except Empty:
            error_message: str = (
                "Internal Server Error - Current state of the state machine is unknown"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def get_mission(self, mission_id: str) -> Mission:
        """Get the mission with mission_id from the current mission planner

        Raises
        ------
        HTTPException 404 Not Found
            If requested mission with mission_id is not found
        HTTPException 500 Internal Server Error
            If for some reason the mission can not be returned
        """
        try:
            return self.mission_planner.get_mission(mission_id)
        except HTTPError as e:
            self.logger.error(e)
            raise HTTPException(status_code=e.response.status_code)
        except MissionNotFoundError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.NOT_FOUND,
                detail=f"Mission with id '{mission_id}' not found",
            )
        except MissionPlannerError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                detail="Could not plan mission",
            )

    def verify_robot_capable_of_mission(
        self, mission: Mission, robot_capabilities: List[str]
    ) -> bool:
        """Verify that the robot is capable of performing this mission

        Raises
        ------
        HTTPException 400 Bad request
            If the robot is not capable of performing mission
        """
        is_capable: bool = True
        missing_capabilities: Set[str] = set()
        for task in mission.tasks:
            if task.type not in robot_capabilities:
                is_capable = False
                missing_capabilities.add(task.type)

        if not is_capable:
            error_message = (
                f"Bad Request - Robot is not capable of performing mission."
                f" Missing functionalities: {missing_capabilities}."
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        return is_capable

    def verify_state_machine_ready_to_receive_mission(self, state: States) -> bool:
        """Verify that the state machine is idle and ready to receive a mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not idle and therefore can not start a new mission
        """
        is_state_machine_ready_to_receive_mission = state == States.Idle
        if not is_state_machine_ready_to_receive_mission:
            error_message = f"Conflict - Robot is not idle - State: {state}"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        return is_state_machine_ready_to_receive_mission

    def start_mission(
        self,
        mission: Mission,
        initial_pose: Optional[Pose],
    ) -> None:
        """Start mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            self._send_command(
                StartMissionMessage(
                    mission=deepcopy(mission),
                    initial_pose=initial_pose,
                ),
                self.queues.start_mission,
            )
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to start mission in ISAR"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Mission started in ISAR")

    def pause_mission(self) -> ControlMissionResponse:
        """Pause mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            return self._send_command(True, self.queues.pause_mission)
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to pause mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        finally:
            self.logger.info("OK - Mission successfully paused")

    def resume_mission(self) -> ControlMissionResponse:
        """Resume mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            return self._send_command(True, self.queues.resume_mission)
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to resume mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        finally:
            self.logger.info("OK - Mission successfully resumed")

    def stop_mission(self) -> ControlMissionResponse:
        """Stop mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            stop_mission_response: ControlMissionResponse = self._send_command(
                True, self.queues.stop_mission
            )
        except QueueTimeoutError:
            error_message = "Internal Server Error - Failed to stop mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        self.logger.info("OK - Mission successfully stopped")
        return stop_mission_response

    def _send_command(self, input: Any, queueio: QueueIO) -> Any:
        queueio.input.put(input)
        try:
            return QueueUtilities.check_queue(
                queueio.output,
                self.queue_timeout,
            )
        except QueueTimeoutError as e:
            QueueUtilities.clear_queue(queueio.input)
            self.logger.error("No output received for command to state machine")
            raise e



from enum import Enum


class MissionStatus(str, Enum):
    NotStarted = "not_started"
    InProgress = "in_progress"
    Paused = "paused"
    Failed = "failed"
    Cancelled = "cancelled"
    Successful = "successful"
    PartiallySuccessful = "partially_successful"


class TaskStatus(str, Enum):
    NotStarted = "not_started"
    InProgress = "in_progress"
    Paused = "paused"
    Failed = "failed"
    Cancelled = "cancelled"
    Successful = "successful"
    PartiallySuccessful = "partially_successful"


class RobotStatus(Enum):
    Available = "available"
    Busy = "busy"
    Offline = "offline"
    Blocked = "blocked"
    BlockedProtectiveStop = "blockedprotectivestop"



from alitra import Frame, Orientation, Pose, Position


class MockPose:
    @staticmethod
    def default_pose():
        return Pose(
            position=Position(x=0, y=0, z=0, frame=Frame("robot")),
            orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame("robot")),
            frame=Frame("robot"),
        )



import json
import os

from alitra import Frame, Orientation, Pose, Position

from isar.apis.models.models import InputOrientation, InputPose, InputPosition
from isar.apis.models.start_mission_definition import (
    InspectionTypes,
    StartMissionDefinition,
    StartMissionInspectionDefinition,
    StartMissionTaskDefinition,
    TaskType,
    to_isar_mission,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import TakeImage


def test_to_isar_mission() -> None:
    DUMMY_MISSION_NAME = "mission_name"

    inspection_definition = StartMissionInspectionDefinition(
        type=InspectionTypes.image,
        inspection_target=InputPosition(x=1, y=1, z=1),
    )
    task_pose = InputPose(
        position=InputPosition(x=1, y=1, z=1),
        orientation=InputOrientation(x=1, y=1, z=1, w=1),
    )
    task_definition = StartMissionTaskDefinition(
        type=TaskType.Inspection,
        pose=task_pose,
        inspection=inspection_definition,
    )
    mission_definition = StartMissionDefinition(
        tasks=[task_definition], name=DUMMY_MISSION_NAME
    )

    isar_mission: Mission = to_isar_mission(mission_definition)

    assert len(isar_mission.id) > 1
    assert isar_mission.name == DUMMY_MISSION_NAME
    assert len(isar_mission.tasks) == 1

    first_task = isar_mission.tasks[0]
    assert len(first_task.id) > 1

    assert isinstance(first_task, TakeImage)

    assert first_task.target == Position(x=1, y=1, z=1, frame=Frame(name="robot"))
    assert len(first_task.inspection_id) > 1
    assert first_task.robot_pose == Pose(
        position=Position(x=1, y=1, z=1, frame=Frame(name="robot")),
        orientation=Orientation(x=1, y=1, z=1, w=1, frame=Frame(name="robot")),
        frame=Frame(name="robot"),
    )


def test_mission_definition_from_json_to_isar_mission() -> None:
    dirname = os.path.dirname(__file__)
    filepath = os.path.join(dirname, "example_mission_definition.json")

    with open(filepath) as f:
        datax = json.load(f)
        mission_definition = StartMissionDefinition(**datax)

    isar_mission: Mission = to_isar_mission(mission_definition)
    assert len(isar_mission.id) > 1
    assert isar_mission.name == "my-mission"

    assert len(isar_mission.tasks) == 1
    task = isar_mission.tasks[0]
    assert len(task.id) > 1
    assert isinstance(task, TakeImage)
    assert task.robot_pose == Pose(
        position=Position(0.0, 0.0, 0.0, frame=Frame("robot")),
        orientation=Orientation(0.0, 0.0, 0.0, 0.0, frame=Frame("robot")),
        frame=Frame("robot"),
    )
    assert task.type == "take_image"
    assert task.target == Position(0.0, 0.0, 0.0, frame=Frame("robot"))



import json
import logging
from pathlib import Path

from injector import inject

from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import (
    MissionNotFoundError,
    MissionPlannerError,
    MissionPlannerInterface,
)
from robot_interface.models.mission.mission import Mission

logger = logging.getLogger("api")


class LocalPlanner(MissionPlannerInterface):
    @inject
    def __init__(self):
        self.predefined_mission_folder = Path(settings.PREDEFINED_MISSIONS_FOLDER)

    def get_mission(self, mission_id) -> Mission:
        missions: dict = self.get_predefined_missions()
        if missions is None:
            raise MissionPlannerError("There were no predefined missions")
        try:
            mission: Mission = missions[mission_id]["mission"]
            return mission
        except KeyError as e:
            raise MissionNotFoundError(
                f"Could not get mission : {mission_id} - does not exist {e}"
            ) from e
        except Exception as e:
            raise MissionPlannerError(f"Could not get mission : {mission_id}") from e

    @staticmethod
    def read_mission_from_file(mission_path: Path) -> Mission:
        with open(mission_path) as json_file:
            mission_dict = json.load(json_file)

        return Mission(**mission_dict)

    def get_predefined_missions(self) -> dict:
        missions: dict = {}
        invalid_mission_ids: list = []
        json_files = self.predefined_mission_folder.glob("*.json")
        for file in json_files:
            mission_name = file.stem
            path_to_file = self.predefined_mission_folder.joinpath(file.name)

            mission: Mission = self.read_mission_from_file(path_to_file)
            if mission.id in invalid_mission_ids:
                logger.warning(
                    f"Duplicate mission id {mission.id} : {path_to_file.as_posix()}"
                )
            elif mission.id in missions:
                conflicting_file_path = missions[mission.id]["file"]
                logger.warning(
                    f"Duplicate mission id {mission.id} : {path_to_file.as_posix()}"
                    + f" and {conflicting_file_path}"
                )
                invalid_mission_ids.append(mission.id)
                missions.pop(mission.id)
            else:
                missions[mission.id] = {
                    "name": mission_name,
                    "file": path_to_file.as_posix(),
                    "mission": mission,
                }
        return missions



import logging
from typing import TYPE_CHECKING

from transitions import State

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Off(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="off", on_enter=self.start)
        self.logger = logging.getLogger("state_machine")
        self.state_machine: "StateMachine" = state_machine

    def start(self):
        self.state_machine.update_state()
        self.logger.info(f"State: {self.state_machine.current_state}")



import logging
from http import HTTPStatus
from typing import Optional

from alitra import Pose
from fastapi import Body, HTTPException, Path
from injector import inject

from isar.apis.models.models import (
    ControlMissionResponse,
    TaskResponse,
    InputPose,
    StartMissionResponse,
)
from isar.apis.models.start_mission_definition import (
    StartMissionDefinition,
    to_isar_mission,
)
from isar.config.settings import robot_settings, settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import (
    TASKS,
    InspectionTask,
    Localize,
    MoveArm,
    ReturnToHome,
)


class SchedulingController:
    @inject
    def __init__(
        self,
        scheduling_utilities: SchedulingUtilities,
    ):
        self.scheduling_utilities: SchedulingUtilities = scheduling_utilities
        self.logger = logging.getLogger("api")

    def start_mission_by_id(
        self,
        mission_id: str = Path(
            alias="id",
            title="Mission ID",
            description="ID-number for predefined mission",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ) -> StartMissionResponse:
        self.logger.info(f"Received request to start mission with id {mission_id}")

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        mission: Mission = self.scheduling_utilities.get_mission(mission_id)
        if return_pose:
            pose: Pose = return_pose.to_alitra_pose()
            mission.tasks.append(ReturnToHome(pose=pose))

        self.scheduling_utilities.verify_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )

        initial_pose_alitra: Optional[Pose] = (
            initial_pose.to_alitra_pose() if initial_pose else None
        )

        self.logger.info(f"Starting mission with ISAR Mission ID: '{mission.id}'")

        self.scheduling_utilities.start_mission(
            mission=mission, initial_pose=initial_pose_alitra
        )

        return self._api_response(mission)

    def start_mission(
        self,
        mission_definition: StartMissionDefinition = Body(
            default=None,
            embed=True,
            title="Mission Definition",
            description="Description of the mission in json format",
        ),
        initial_pose: Optional[InputPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial "
            "localization of robot",
            embed=True,
        ),
        return_pose: Optional[InputPose] = Body(
            default=None,
            description="End pose of the mission. The robot return to the specified "
            "pose after finishing all inspections",
            embed=True,
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new mission")

        if not mission_definition:
            error_message: str = (
                "Unprocessable entity - 'mission_definition' empty or invalid"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY, detail=error_message
            )

        state: States = self.scheduling_utilities.get_state()
        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        try:
            mission: Mission = to_isar_mission(
                start_mission_definition=mission_definition, return_pose=return_pose
            )
        except MissionPlannerError as e:
            error_message = f"Bad Request - Cannot create ISAR mission: {e}"
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        self.scheduling_utilities.verify_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )

        initial_pose_alitra: Optional[Pose] = (
            initial_pose.to_alitra_pose() if initial_pose else None
        )

        self.logger.info(f"Starting mission: {mission.id}")
        self.scheduling_utilities.start_mission(
            mission=mission, initial_pose=initial_pose_alitra
        )
        return self._api_response(mission)

    def pause_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to pause current mission")

        state: States = self.scheduling_utilities.get_state()

        if state not in [
            States.Monitor,
            States.Initiate,
        ]:
            error_message = (
                f"Conflict - Pause command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.CONFLICT,
                detail=error_message,
            )

        pause_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.pause_mission()
        )
        return pause_mission_response

    def resume_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to resume current mission")

        state: States = self.scheduling_utilities.get_state()

        if state != States.Paused:
            error_message = (
                f"Conflict - Resume command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        resume_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.resume_mission()
        )
        return resume_mission_response

    def stop_mission(self) -> ControlMissionResponse:
        self.logger.info("Received request to stop current mission")

        state: States = self.scheduling_utilities.get_state()

        if state == States.Off:
            error_message = (
                f"Conflict - Stop command received in invalid state - State: {state}"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

        stop_mission_response: ControlMissionResponse = (
            self.scheduling_utilities.stop_mission()
        )
        return stop_mission_response

    def drive_to(
        self,
        target_pose: InputPose = Body(
            default=None,
            title="Target Pose",
            description="The target pose for the drive_to task",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new drive-to mission")

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        pose: Pose = target_pose.to_alitra_pose()
        mission: Mission = Mission(
            name="Drive to pose", tasks=[ReturnToHome(pose=pose)]
        )

        self.logger.info(
            f"Starting drive to mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(mission=mission, initial_pose=None)
        return self._api_response(mission)

    def start_localization_mission(
        self,
        localization_pose: InputPose = Body(
            default=None,
            embed=True,
            title="Localization Pose",
            description="The current position of the robot",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new localization mission")

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        pose: Pose = localization_pose.to_alitra_pose()
        mission: Mission = Mission(
            name="Localization mission", tasks=[Localize(localization_pose=pose)]
        )

        self.logger.info(
            f"Starting localization mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(
            mission=mission,
            initial_pose=None,
        )
        return self._api_response(mission)

    def start_move_arm_mission(
        self,
        arm_pose_literal: str = Path(
            ...,
            alias="arm_pose_literal",
            title="Arm pose literal",
            description="Arm pose as a literal",
        ),
    ) -> StartMissionResponse:
        self.logger.info("Received request to start new move arm mission")

        if not robot_settings.VALID_ARM_POSES:
            error_message: str = (
                f"Received a request to move the arm but the robot "
                f"{settings.ROBOT_NAME} does not support moving an arm"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST, detail=error_message
            )

        if arm_pose_literal not in robot_settings.VALID_ARM_POSES:
            error_message = (
                f"Received a request to move the arm but the arm pose "
                f"{arm_pose_literal} is not supported by the robot "
                f"{settings.ROBOT_NAME}"
            )
            self.logger.warning(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST, detail=error_message
            )

        state: States = self.scheduling_utilities.get_state()

        self.scheduling_utilities.verify_state_machine_ready_to_receive_mission(state)

        mission: Mission = Mission(
            name="Move arm mission", tasks=[MoveArm(arm_pose=arm_pose_literal)]
        )

        self.logger.info(
            f"Starting move arm mission with ISAR Mission ID: '{mission.id}'"
        )
        self.scheduling_utilities.start_mission(
            mission=mission,
            initial_pose=None,
        )
        return self._api_response(mission)

    def _api_response(self, mission: Mission) -> StartMissionResponse:
        return StartMissionResponse(
            id=mission.id,
            tasks=[self._task_api_response(task) for task in mission.tasks],
        )

    def _task_api_response(self, task: TASKS) -> TaskResponse:
        if isinstance(task, InspectionTask):
            inspection_id = task.inspection_id
        else:
            inspection_id = None

        return TaskResponse(
            id=task.id, tag_id=task.tag_id, inspection_id=inspection_id, type=task.type
        )



from typing import List, Optional

from alitra import Frame, Orientation, Pose, Position
from pydantic import BaseModel, Field

from robot_interface.models.mission.task import TaskTypes


class TaskResponse(BaseModel):
    id: str
    tag_id: Optional[str] = None
    inspection_id: Optional[str] = None
    type: TaskTypes


class StartMissionResponse(BaseModel):
    id: str
    tasks: List[TaskResponse]


class ControlMissionResponse(BaseModel):
    mission_id: str
    mission_status: str
    task_id: str
    task_status: str


class RobotInfoResponse(BaseModel):
    robot_package: str
    isar_id: str
    robot_name: str
    robot_capabilities: List[str]
    robot_map_name: str
    plant_short_name: str


class InputOrientation(BaseModel):
    x: float
    y: float
    z: float
    w: float
    frame_name: str = Field(default="robot")

    def to_alitra_orientation(self) -> Orientation:
        return Orientation(
            x=self.x,
            y=self.y,
            z=self.z,
            w=self.w,
            frame=Frame(self.frame_name),
        )


class InputPosition(BaseModel):
    x: float
    y: float
    z: float
    frame_name: str = Field(default="robot")

    def to_alitra_position(self) -> Position:
        return Position(
            x=self.x,
            y=self.y,
            z=self.z,
            frame=Frame(self.frame_name),
        )


class InputPose(BaseModel):
    position: InputPosition
    orientation: InputOrientation
    frame_name: str = Field(default="robot")

    def to_alitra_pose(self) -> Pose:
        return Pose(
            position=self.position.to_alitra_position(),
            orientation=self.orientation.to_alitra_orientation(),
            frame=Frame(self.frame_name),
        )
